"""Loading weather data (temp)
"""
import os
import numpy as np

def get_all_station_per_weather_yr(path_to_csvs, min_nr_of_stations=10):
    """Get all weather years and stations

    Arguments
    -----------
    path_to_csvs : path
        Path to temperature data
    min_nr_of_stations : int,default=10
        Number of stations which are needed in a year
        in order that year is used

    Returns
    ---------
    cleaned_out_dict : dict
        All years and stations which fulfill minimum
        number of stations criteria
    """
    out_dict = {}

    all_txt_files_in_folder = os.listdir(path_to_csvs)

    for file_path in all_txt_files_in_folder:
        year = int(file_path.split("__")[0])
        weather_station_nr = int(file_path.split("__")[1][:-4])

        try:
            out_dict[year].append(weather_station_nr)
        except KeyError:
            out_dict[year] = [weather_station_nr]

    # Remove all years with lower number of stations
    cleaned_out_dict = {}
    for year, stations in out_dict.items():
        if len(stations) >= min_nr_of_stations:
            cleaned_out_dict[year] = stations
        else:
            print("The year {} has not enough stations".format(year))

    print("Total number of years: " + str(len(cleaned_out_dict.keys())))
    return cleaned_out_dict

def read_weather_data_script_data(path_to_csv, weather_yrs_scenario):
    """Read in weather data from script data

    Arguments
    ----------
    path_to_csv : str
        Path

    Returns
    -------
    temp_data : dict
        Temperature yh per weater station
    """
    temp_data = {}

    all_txt_files_in_folder = os.listdir(path_to_csv)

    for file_path in all_txt_files_in_folder:
        path_file_to_read = os.path.join(path_to_csv, file_path)
        year = int(file_path.split("__")[0])
        station_id = int(file_path.split("__")[1][:-4]) #remove .txt

        if year == weather_yrs_scenario:
            temp_data[station_id] = np.loadtxt(path_file_to_read, delimiter=',')
        else:
            pass

    return temp_data



"""
"""
import numpy as np
from energy_demand.technologies import fuel_service_switch
from energy_demand.read_write import read_data

def test_get_fuel_switches_enduse():

    service_switches = {'regA': [
        read_data.ServiceSwitch(
            enduse='heating',
            sector=None,
            technology_install='techA',
            service_share_ey=0.5,
            switch_yr=2020),
        read_data.ServiceSwitch(
            enduse='other',
            sector=None,
            technology_install='techB',
            service_share_ey=0.5,
            switch_yr=2020)]}

    out = fuel_service_switch.get_switches_of_enduse(
        switches=service_switches,
        enduse='heating')

    for reg, switches in out.items():
        for switch in switches:
            assert switch.enduse == 'heating'

    # --
    service_switches = {'regA': [
        read_data.ServiceSwitch(
            enduse='heating',
            sector=None,
            technology_install='techA',
            service_share_ey=0.5,
            switch_yr=2020),
        read_data.ServiceSwitch(
            enduse='other',
            sector=None,
            technology_install='techB',
            service_share_ey=0.5,
            switch_yr=2020)]}

    out = fuel_service_switch.get_switches_of_enduse(
        switches=service_switches,
        enduse='heating')

    for reg in out:
        for switch in out[reg]:
            assert switch.enduse == 'heating'

def test_switches_to_dict():
    """testing"""
    service_switches = {'regA': [
        read_data.ServiceSwitch(
            enduse='heating',
            sector=None,
            technology_install='techA',
            service_share_ey=0.5,
            switch_yr=2020),
        read_data.ServiceSwitch(
            enduse='heating',
            sector=None,
            technology_install='techB',
            service_share_ey=0.5,
            switch_yr=2020)]}

    out = fuel_service_switch.switches_to_dict(
        service_switches=service_switches)

    assert out['regA'] == {'techA': 0.5, 'techB': 0.5}

def test_get_switch_criteria():
    """testing
    """
    crit_switch_happening = {'heating': ['sectorA']}

    crit = fuel_service_switch.get_switch_criteria(
        enduse='heating',
        sector='sectorA',
        crit_switch_happening=crit_switch_happening,
        base_yr=2015,
        curr_yr=2015)

    assert crit == False

    crit = fuel_service_switch.get_switch_criteria(
        enduse='heating',
        sector='sectorA',
        crit_switch_happening=crit_switch_happening,
        base_yr=2015,
        curr_yr=2025)

    assert crit == True

    crit = fuel_service_switch.get_switch_criteria(
        enduse='heating',
        sector='sectorB',
        crit_switch_happening=crit_switch_happening,
        base_yr=2015,
        curr_yr=2025)

    assert crit == False

def test_sum_fuel_across_sectors():
    """
    """
    fuels = {'sectorA': np.ones(10), 'sectorB': np.ones(10)}
    sum = fuel_service_switch.sum_fuel_across_sectors(fuels)

    assert np.sum(sum) == 20

def test_create_switches_from_s_shares():

    enduse_switches = [read_data.ServiceSwitch(
        enduse='heating',
        technology_install='techA',
        switch_yr=2050,
        service_share_ey=0.6)]

    out = fuel_service_switch.create_switches_from_s_shares(
        enduse='heating',
        s_tech_by_p= {None:{'heating': {'techA': 0.2, 'techB': 0.8}}},
        switch_technologies=['techA'],
        tech_enduse_by=['techA', 'techB'],
        enduse_switches=enduse_switches,
        s_tot_defined=0.6,
        sector=None,
        switch_yr=2050)

    for switch in out:
        if switch.technology_install == 'techA':
            assert switch.service_share_ey == 0.6
        if switch.technology_install == 'techB':
            assert switch.service_share_ey == 0.4

def test_autocomplete_switches():
    """testing"""
    service_switches =  [read_data.ServiceSwitch(
        enduse='heating',
        technology_install='techA',
        switch_yr=2050,
        service_share_ey=0.6)]
    defined_temporal_narrative_points = {'heating': [2050]}

    enduses = {'submodelA': ['heating']}
    sectors = {'submodelA': [None]}

    out = fuel_service_switch.autocomplete_switches(
        service_switches=service_switches,
        specified_tech_enduse_by={'heating': {None: ['techA', 'techB', 'techC']}},
        s_tech_by_p={None: {'heating': {'techA': 0.2, 'techB': 0.4, 'techC': 0.4}}},
        crit_all_the_same=True,
        enduses=enduses,
        sectors=sectors,
        regions=['regA'],
        f_diffusion=False,
        techs_affected_spatial_f=False,
        service_switches_from_capacity={'regA':[]})

    assert out[None]['heating']['regA'][2050]['techA'] == 0.6
    assert out[None]['heating']['regA'][2050]['techB'] == 0.2
    assert out[None]['heating']['regA'][2050]['techC'] == 0.2

    # ---
    service_switches = [read_data.ServiceSwitch(
        enduse='heating',
        technology_install='techA',
        switch_yr=2050,
        service_share_ey=0.6)]

    out = fuel_service_switch.autocomplete_switches(
        service_switches=service_switches,
        specified_tech_enduse_by={'heating': {None: ['techA', 'techB', 'techC']}},
        s_tech_by_p= {None: {'heating': {'techA': 0.2, 'techB': 0.4, 'techC': 0.4}}},
        crit_all_the_same=True,
        enduses=enduses,
        sectors=sectors,
        regions=['regA'],
        f_diffusion={'heating':{'regA':1.0, 'regB': 1.0}},
        techs_affected_spatial_f=['techA'],
        service_switches_from_capacity={'regA': [], 'regB': []})

    assert out[None]['heating']['regA'][2050]['techA'] == 0.6
    assert out[None]['heating']['regA'][2050]['techB'] == 0.2
    assert out[None]['heating']['regA'][2050]['techC'] == 0.2

def test_create_service_switch():
    """testing
    """
    installed_capacity = 100

    enduse = 'heating'

    capacity_switches = [read_data.CapacitySwitch(
        enduse='heating',
        technology_install='boiler_gas',
        switch_yr=2020,
        installed_capacity=installed_capacity)]

    capacity_switch = read_data.CapacitySwitch(
        enduse='heating',
        technology_install='boiler_gas',
        switch_yr=2020,
        installed_capacity=installed_capacity)

    assumptions = {
        'technologies':
            {'boiler_oil': read_data.TechnologyData(
                fueltype='oil',
                eff_by=1.0,
                eff_ey=1.0,
                year_eff_ey=2020,
                eff_achieved=1.0,
                diff_method='linear',
                market_entry=2010,
                tech_type='tech_heating',
                tech_max_share=1.0),
            'boiler_gas': read_data.TechnologyData(
                fueltype='gas',
                eff_by=1.0,
                eff_ey=1.0,
                year_eff_ey=2020,
                eff_achieved=1.0,
                diff_method='linear',
                market_entry=2010,
                tech_type='tech_heating',
                tech_max_share=1.0)
            },
        'rs_fuel_tech_p_by': {0: {'boiler_gas': 0.0}, 1: {'boiler_oil': 1.0}}
    }

    base_yr = 2015

    fuels = {0: 0, 1: 10} #array originally
    sector = None
    narrative_timesteps = [2020]

    results = fuel_service_switch.create_service_switch(
        narrative_timesteps,
        enduse,
        sector,
        capacity_switch,
        capacity_switches,
        assumptions['technologies'],
        assumptions['rs_fuel_tech_p_by'],
        base_yr,
        fuels)

    # Fuel share boiler_gas
    expected = 1 / (10 + installed_capacity) * installed_capacity

    for entry in results:
        if entry.technology_install == 'boiler_gas':
            assert expected == entry.service_share_ey

def test_capacity_switch():
    """Testing
    """
    technologies = {'techA': read_data.TechnologyData(
        fueltype='oil',
        eff_by=1.0,
        eff_ey=1.0,
        year_eff_ey=2020,
        eff_achieved=1.0,
        diff_method='linear',
        market_entry=2010,
        tech_type='tech_heating',
        tech_max_share=1.0)}

    capacity_switches ={"regA": [
        read_data.CapacitySwitch(
            enduse='heating',
            technology_install='techA',
            switch_yr=2050,
            installed_capacity=200)]}

    narrative_timesteps = {'heating': [2050]}
    regions = ["regA"]
    result_service_switches = fuel_service_switch.capacity_switch(
        narrative_timesteps=narrative_timesteps,
        regions=regions,
        capacity_switches=capacity_switches,
        technologies=technologies,
        fuels={'heating': {0: 100}},
        fuel_shares_enduse_by={'heating': {0: {'techA': 1.0}}},
        base_yr=2015)

    for region in regions:
        for switch in result_service_switches[region]:
            if switch.technology_install == 'techA':
                assert switch.service_share_ey == 1.0

    #---------------

    technologies = {
        'techA': read_data.TechnologyData(
            fueltype='oil',
            eff_by=1.0,
            eff_ey=1.0,
            year_eff_ey=2020,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=2010,
            tech_type='tech_heating',
            tech_max_share=1.0),
        'techB': read_data.TechnologyData(
            fueltype='oil',
            eff_by=1.0,
            eff_ey=1.0,
            year_eff_ey=2020,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=2010,
            tech_type='tech_heating',
            tech_max_share=1.0)}

    capacity_switches = {"regA":[
        read_data.CapacitySwitch(
            enduse='heating',
            technology_install='techA',
            switch_yr=2050,
            installed_capacity=200)
    ]}

    result_service_switches = fuel_service_switch.capacity_switch(
        narrative_timesteps=narrative_timesteps,
        regions=regions,
        capacity_switches=capacity_switches,
        technologies=technologies,
        fuels={'heating': {0: 100}},
        fuel_shares_enduse_by={'heating': {0: {'techA': 0.5, 'techB': 0.5}}},
        base_yr=2015)

    for region in regions:
        for switch in result_service_switches[region]:
            if switch.technology_install == 'techA':
                assert round(switch.service_share_ey, 3) == round((1 / (300)) * 250, 3)
            if switch.technology_install == 'techB':
                assert round(switch.service_share_ey, 3) == round((1 / (300)) * 50, 3)
test_capacity_switch()
def autocomplete_switches():
    """Testing
    """
    service_switches = [read_data.ServiceSwitch(
        'heating',
        'techA',
        0.5,
        2020)]

    specified_tech_enduse_by = {'heating': ['techA', 'techB']}
    s_tech_by_p = {'heating': {'techA': 0.8, 'techB': 0.2}}

    service_switches = fuel_service_switch.autocomplete_switches(
        service_switches=service_switches,
        specified_tech_enduse_by=specified_tech_enduse_by,
        s_tech_by_p=s_tech_by_p)
    
    for switch in service_switches:
        if switch.technology_install == 'techA':
            assert switch.service_share_ey == 0.5
        
        if switch.technology_install == 'techB':
            assert switch.service_share_ey == 0.5
    #---

    service_switches = [read_data.ServiceSwitch(
        'heating',
        'techA',
        0.3,
        2020)]

    specified_tech_enduse_by = {'heating': ['techA', 'techB', 'techC']}
    s_tech_by_p = {'heating': {'techA': 0.6, 'techB': 0.2,'techC': 0.1}}

    service_switches = fuel_service_switch.autocomplete_switches(
        service_switches=service_switches,
        specified_tech_enduse_by=specified_tech_enduse_by,
        s_tech_by_p=s_tech_by_p)

    for switch in service_switches:
        if switch.technology_install == 'techA':
            assert switch.service_share_ey == 0.3
        if switch.technology_install == 'techB':
            assert switch.service_share_ey == 0.7 * (2.0 / 3.0)
        if switch.technology_install == 'techC':
            assert switch.service_share_ey == 0.7 * (1.0 / 3.0)

def test_get_share_s_tech_ey():
    """testing"""

    service_switches = [read_data.ServiceSwitch(
        enduse='heating',
        sector=None,
        technology_install='techA',
        service_share_ey=0.3,
        switch_yr=2020)]

    service_switches = {'regA': [read_data.ServiceSwitch(
        enduse='heating',
        sector=None,
        technology_install='techA',
        service_share_ey=0.3,
        switch_yr=2020)]}

    defined_temporal_narrative_points = [2020] 
    result = fuel_service_switch.get_share_s_tech_ey(
        service_switches=service_switches)

    assert result['heating']['regA'][2020]['techA'] == 0.3



"""Read residential raw files and store to txt
"""
import os
import csv
from datetime import date
import logging
import numpy as np
from energy_demand.basic import date_prop
from energy_demand.read_write import read_data
from energy_demand.read_write import write_data
from energy_demand.profiles import load_profile as lp

def read_csv(path_to_csv):
    """This function reads in CSV files and skips header row.

    Arguments
    ----------
    path_to_csv : str
        Path to csv file
    _dt : str
        Defines dtype of array to be read in (takes float)

    Returns
    -------
    elements_array : array_like
        Returns an array `elements_array` with the read in csv files.

    Notes
    -----
    The header row is always skipped.
    """
    service_switches = []
    with open(path_to_csv, 'r') as csvfile:
        read_lines = csv.reader(csvfile, delimiter=',')
        _headings = next(read_lines) # Skip first row

        # Iterate rows
        for row in read_lines:
            service_switches.append(row)

    return np.array(service_switches) # Convert list into array

def get_hes_load_shapes(
        appliances_hes_matching,
        year_raw_values,
        hes_y_peak,
        enduse,
        single_enduse=True,
        enduses=False
    ):
    """Read in raw HES data and generate shapes

    Calculate peak day demand

    Arguments
    ----------
    appliances_hes_matching : dict
        HES appliances are matched witch enduses
    year_raw_values : array
        Yearly values from raw
    hes_y_peak : data
        Peak raw values
    enduse : string
        enduse

    Returns
    -------
    shape_peak_yh : float
        Peak demand of each hours of peak day

    Notes
    -----
    The HES appliances are matched with enduses
    """
    if single_enduse:
        # Match enduse with HES appliance ID
        # (see look_up table in original files for more information)
        hes_app_id = appliances_hes_matching[enduse]

        # Values
        enduse_values = year_raw_values[:, :, hes_app_id]

        # Total yearly demand of hes_app_id
        tot_enduse_y = np.sum(year_raw_values[:, :, hes_app_id])

        # Get peak daily load shape, calculate amount of energy demand for peak day of hes_app_id
        peak_h_values = hes_y_peak[:, hes_app_id]

    else:
        for multi_enduse in enduses:

            # Id of enduse
            hes_app_id = appliances_hes_matching[multi_enduse]

            try:
                tot_enduse_y += np.sum(year_raw_values[:, :, hes_app_id])

                # Get peak daily load shape, calculate amount of energy demand for peak day of hes_app_id
                peak_h_values += hes_y_peak[:, hes_app_id]

                # Values
                enduse_values += year_raw_values[:, :, hes_app_id]

            except:
                # Total yearly demand of hes_app_id
                tot_enduse_y = np.sum(year_raw_values[:, :, hes_app_id])
                
                # Get peak daily load shape, calculate amount of energy demand for peak day of hes_app_id
                peak_h_values = hes_y_peak[:, hes_app_id]

                enduse_values = year_raw_values[:, :, hes_app_id]

    # Maximum daily demand
    tot_peak_demand_d = np.sum(peak_h_values)

    # Shape of peak day (hourly values of peak day)
    shape_peak_dh = lp.abs_to_rel(peak_h_values)

    # ---Calculate non-peak shapes
    shape_non_peak_yd = np.zeros((365), dtype="float")
    shape_non_peak_y_dh = np.zeros((365, 24), dtype="float")

    for day, day_values in enumerate(enduse_values):
        #day_values = year_raw_values[day, :, hes_app_id]

        shape_non_peak_yd[day] = (1.0 / tot_enduse_y) * np.sum(day_values)
        shape_non_peak_y_dh[day] = (1.0 / np.sum(day_values)) * day_values

    return shape_peak_dh, shape_non_peak_y_dh, shape_non_peak_yd

def assign_hes_data_to_year(nr_of_appliances, hes_data, base_yr):
    """Fill every base year day with correct data

    Arguments
    ----------
    nr_of_appliances : dict
        Defines how many appliance types are stored (max 10 provided in original hes file)
    hes_data : array
        HES raw data for every month and daytype and appliance
    base_yr : float
        Base year to generate shapes

    Returns
    -------
    year_raw_values : array
        Energy data for every day in the base year for every appliances
    """
    year_raw_values = np.zeros((365, 24, nr_of_appliances), dtype="float") #yeardays, houry, appliances

    # Create list with all dates of a whole year
    list_dates = date_prop.fullyear_dates(
        start=date(base_yr, 1, 1),
        end=date(base_yr, 12, 31))

    # Assign every date to the place in the array of the year
    for yearday_date in list_dates:

        month_python = date_prop.get_month_from_yeraday(
            yearday_date.timetuple().tm_year,
            yearday_date.timetuple().tm_yday)

        yearday_python = date_prop.date_to_yearday(
            yearday_date.timetuple().tm_year,
            yearday_date.timetuple().tm_mon,
            yearday_date.timetuple().tm_mday)

        daytype_str = date_prop.get_weekday_type(yearday_date)

        # Get day from HES raw data array
        year_raw_values[yearday_python] = hes_data[daytype_str][month_python]

    return year_raw_values

def read_hes_data(paths_hes, nr_app_type_lu):
    """Read in HES raw csv files and provide for every
    day in a year (yearday) all fuels

    The fuel is provided for every daytype (weekend or working day)
    for every month and every appliance_typ

    Arguments
    ----------
    paths_hes : string
        Path to HES raw data file
    nr_app_type_lu : dict
        Number of appliances (defines size of container to store data)

    Returns
    -------
    hes_data : dict
        HES non peak raw data per fueltype
    hes_y_coldest : array
        HES for coldest day

    Note
    ----
        -   As only shapes are generated, the absolute
            values are irrelevant, i.e. the unit of energy
    """
    day_type_lu = {
        0: 'holiday',
        1: 'working_day',
        2: 'coldest'}

    hes_data = {}
    for day_type_str in day_type_lu.values():
        hes_data[day_type_str] = np.zeros((12, 24, nr_app_type_lu), dtype="float")

    hes_y_coldest = np.zeros((24, nr_app_type_lu), dtype="float")

    # Read in raw HES data from CSV
    raw_elec_data = read_csv(paths_hes)

    # Iterate raw data of hourly eletrictiy demand
    for row in raw_elec_data:
        month_int, day_type, appliance_typ = int(row[0]), str(row[1]), int(row[2])
        k_header = 3 #Row in Excel where energy data start

        for hour in range(24):
            hourly_value = float(row[k_header])

            # if coldest (see HES file)
            if day_type == 'coldest':
                hes_y_coldest[hour][appliance_typ] = hourly_value
                k_header += 1
            else:
                hes_data[day_type][month_int][hour][appliance_typ] = hourly_value
                k_header += 1

    return hes_data, hes_y_coldest

def run(paths, local_paths, base_yr):
    """Function to run script
    """
    print("... start script %s", os.path.basename(__file__))

    hes_appliances_matching = {
        'rs_cold': 0,
        'rs_cooking': 1,
        'rs_lighting': 2,
        'rs_consumer_electronics': 3,
        'rs_home_computing': 4,
        'rs_wet': 5,
        'rs_water_heating_water_heating': 6,
        'NOT_USED_unkown_1': 7,
        'NOT_USED_unkown_2': 8,
        'NOT_USED_unkown_3': 9,
        'rs_water_heating_showers': 10}

    # HES data -- Generate generic load profiles for all electricity appliances from HES data
    hes_data, hes_y_peak = read_hes_data(
        paths['lp_rs'],
        len(hes_appliances_matching))

    # Assign read in raw data to the base year
    year_raw_hes_values = assign_hes_data_to_year(
        len(hes_appliances_matching),
        hes_data,
        int(base_yr))

    rs_enduses, _, _ = read_data.read_fuel_rs(
        paths['rs_fuel_raw'])

    # Load shape for all enduses
    for enduse in rs_enduses:
        for dummy_sector in rs_enduses[enduse]:
            if enduse not in hes_appliances_matching:

                # Merge shapes of water heating and shower
                if enduse == 'rs_water_heating':

                    # Generate HES load shapes
                    shape_peak_dh, shape_non_peak_y_dh, shape_non_peak_yd = get_hes_load_shapes(
                        hes_appliances_matching,
                        year_raw_hes_values,
                        hes_y_peak,
                        enduse,
                        single_enduse=False,
                        enduses=['rs_water_heating_showers', 'rs_water_heating_water_heating'])

                    # Write txt files
                    write_data.create_txt_shapes(
                        enduse,
                        local_paths['rs_load_profile_txt'],
                        shape_peak_dh,
                        shape_non_peak_y_dh,
                        shape_non_peak_yd)
                else:
                    pass
            else:
                # Generate HES load shapes
                shape_peak_dh, shape_non_peak_y_dh, shape_non_peak_yd = get_hes_load_shapes(
                    hes_appliances_matching,
                    year_raw_hes_values,
                    hes_y_peak,
                    enduse)

                # Write txt files
                write_data.create_txt_shapes(
                    enduse,
                    local_paths['rs_load_profile_txt'],
                    shape_peak_dh,
                    shape_non_peak_y_dh,
                    shape_non_peak_yd)

    logging.info("... finished script %s", os.path.basename(__file__))
    return



"""Plot stacked enduses for each submodel
"""
import logging
import numpy as np
import matplotlib.pyplot as plt

from energy_demand.basic import conversions
from energy_demand.plotting import basic_plot_functions

def run(
        years_simulated,
        results_enduse_every_year,
        enduses,
        color_list,
        fig_name,
        plot_legend=True
    ):
    """Plots stacked energy demand

    Arguments
    ----------
    years_simulated : list
        Simulated years
    results_enduse_every_year : dict
        Results [year][enduse][fueltype_array_position]

    enduses :
    fig_name : str
        Figure name

    Note
    ----
        -   Sum across all fueltypes
        -   Not possible to plot single year

    https://matplotlib.org/examples/pylab_examples/stackplot_demo.html
    """
    print("...plot stacked enduses")
    nr_of_modelled_years = len(years_simulated)

    x_data = np.array(years_simulated)

    y_value_arrays = []
    legend_entries = []

    for enduse in enduses:
        legend_entries.append(enduse)

        y_values_enduse_yrs = np.zeros((nr_of_modelled_years))

        for year_array_nr, model_year in enumerate(results_enduse_every_year.keys()):

            # Sum across all fueltypes
            tot_across_fueltypes = np.sum(results_enduse_every_year[model_year][enduse])

            # Conversion: Convert GWh per years to GW
            yearly_sum_twh = conversions.gwh_to_twh(tot_across_fueltypes)

            logging.debug("... model_year {} enduse {}  twh {}".format(
                model_year, enduse, np.sum(yearly_sum_twh)))

            if yearly_sum_twh < 0:
                raise Exception("no minus values allowed {}  {}  {}".format(enduse, yearly_sum_twh, model_year))

            y_values_enduse_yrs[year_array_nr] = yearly_sum_twh

        # Add array with values for every year to list
        y_value_arrays.append(y_values_enduse_yrs)

    # Try smoothing line
    try:
        x_data_smoothed, y_value_arrays_smoothed = basic_plot_functions.smooth_data(
            x_data, y_value_arrays, num=40000)
    except:
        x_data_smoothed = x_data
        y_value_arrays_smoothed = y_value_arrays

    # Convert to stacked
    y_stacked = np.row_stack((y_value_arrays_smoothed))

    # Set figure size
    fig = plt.figure(
        figsize=basic_plot_functions.cm2inch(8, 8))

    ax = fig.add_subplot(1, 1, 1)

    # ----------
    # Stack plot
    # ----------
    color_stackplots = color_list[:len(enduses)]

    ax.stackplot(
        x_data_smoothed,
        y_stacked,
        alpha=0.8,
        colors=color_stackplots)

    if plot_legend:
        plt.legend(
            legend_entries,
            prop={
                'family':'arial',
                'size': 5},
            ncol=2,
            loc='upper center',
            bbox_to_anchor=(0.5, -0.1),
            frameon=False,
            shadow=True)

    # -------
    # Axis
    # -------
    year_interval = 10
    major_ticks = np.arange(
        years_simulated[0],
        years_simulated[-1] + year_interval,
        year_interval)

    plt.xticks(major_ticks, major_ticks)

    plt.ylim(ymax=500)
    #yticks = [100, 200, 300, 400, 500]
    #plt.yticks(yticks, yticks)
    # -------
    # Labels
    # -------
    plt.ylabel("TWh", fontsize=10)
    plt.xlabel("Year", fontsize=10)
    #plt.title("ED whole UK", fontsize=10)

    # Tight layout
    fig.tight_layout()

    plt.margins(x=0)
    plt.savefig(fig_name)
    plt.close()



"""Clean original temperature data
"""
import os
import csv
import collections
import numpy as np

from energy_demand.basic import date_prop
from energy_demand.basic import basic_functions

def count_sequence_of_zeros(sequence):
    cnt = 0
    max_cnt_zeros = 0
    for hour_val in sequence:
        if hour_val == 0:
            cnt += 1
        else:
            if cnt > max_cnt_zeros:
                max_cnt_zeros = cnt
            cnt = 0 #reset to zero again
    
    return max_cnt_zeros

def nan_helper(y):
    """Helper to handle indices and logical indices of NaNs.
    Input:
        - y, 1d numpy array with possible NaNs
    Output:
        - nans, logical indices of NaNs
        - index, a function, with signature indices= index(logical_indices),
          to convert logical indices of NaNs to 'equivalent' indices
    Example:
        >>> # linear interpolation of NaNs
        >>> nans, x= nan_helper(y)
        >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])
    Taken from: https://stackoverflow.com/questions/6518811/interpolate-nan-values-in-a-numpy-array
    """
    return np.isnan(y), lambda z: z.nonzero()[0]

def run(
        path_files,
        path_out_files,
        crit_missing_values=100,
        crit_nr_of_zeros=500,
        nr_daily_zeros=10
    ):
    """Iterate weather data from MIDAS and
    generate annual hourly temperatre data files
    for every weather station by interpolating missing values
    http://data.ceda.ac.uk/badc/ukmo-midas/data/WH/yearly_files/
    The data are obtained from the Centre for Environmental Data Analysis
    [Download link]http://data.ceda.ac.uk/badc/ukmo-midas/data/WH/yearly_files/ ()
    http://badc.nerc.ac.uk/artefacts/badc_datadocs/ukmo-midas/WH_Table.html (metadata)
    Weather Stations information: http://badc.nerc.ac.uk/search/midas_stations/
    http://badc.nerc.ac.uk/cgi-bin/midas_stations/search_by_name.cgi.py?name=&minyear=&maxyear=
    The values are then written to a created folder as follows:
        year__stationname.txt
    Arguments
    ---------
    path_files : str
        Path to folder with original weather data files
    path_out_files : str
        Path to folder where the cleaned data are stored
    crit_missing_values : int
        Criteria of how many missing values there mus bet until
        a weather station is discarded
    crit_nr_of_zeros : int
        Criteria of how many zeros there must be per year until
        weather station is discarde
    nr_daily_zeros : int
        How many zero values there can be maxmum in a day in the
        year until the station is discarded
    Note
    -----
        - In case of a leap year the 29 of February is ignored
    """
    print("... starting to clean original weather files")

    # Stations which are outisde of the uk and are ignored
    stations_outside_UK = [
        1605, # St. Helena
        1585, # Gibraltar
        1609] # Falkland Islands

    # Create out folder to store station specific csv
    if os.path.isdir(path_out_files):
        basic_functions.delete_folder(path_out_files)
        os.mkdir(path_out_files)
    else:
        os.mkdir(path_out_files)

    # Placeholder value for missing entry
    placeholder_value = np.nan

    # Read all files
    all_annual_raw_files = os.listdir(path_files)

    # Sort according to year
    all_annual_raw_files.sort()

    # Annual temperature file
    for file_name in all_annual_raw_files:
        print(" ")
        print("... reading csv file: " + str(file_name), flush=True)
        print(" ")
        path_to_csv = os.path.join(path_files, file_name)

        temp_stations = {}

        with open(path_to_csv, 'r') as csvfile:
            read_lines = csv.reader(csvfile, delimiter=',')

            for row in read_lines:

                date_measurement = row[0].split(" ")
                year = int(date_measurement[0].split("-")[0])
                month = int(date_measurement[0].split("-")[1])
                day = int(date_measurement[0].split("-")[2])
                hour = int(date_measurement[1][:2])

                yearday = date_prop.date_to_yearday(year, month, day)

                if date_prop.is_leap_year(year):
                    yearday = yearday - 1 # Substract because 29. of Feb is later ignored
                else:
                    pass

                year_hour = (yearday * 24) + hour

                # Weather station id
                station_id = int(row[5])

                # If station is outside uk or leap year day
                if (station_id in stations_outside_UK) or (month == 2 and day == 29):
                    pass
                else:
                    # Air temperature in Degrees Celcius
                    if row[35] == ' ' or row[35] == '': # If no data point
                        air_temp = placeholder_value
                    else:
                        air_temp = float(row[35])

                    # Add weather station if not already added to dict
                    if station_id not in temp_stations:
                        temp_stations[station_id] = np.zeros((8760), dtype="float")
                    else:
                        pass

                    temp_stations[station_id][year_hour] = air_temp

        # ------------------------------------------
        # Interpolate missing values (np.nan)
        # ------------------------------------------
        temp_stations_cleaned_reshaped = {}

        for station in list(temp_stations.keys()):

            # ------------------------
            # Number of empty  values
            # ------------------------
            nans, x = nan_helper(temp_stations[station])
            nr_of_nans = list(nans).count(True)

            # ------------------------
            # nr of zeros
            # ------------------------
            try:
                nr_of_zeros = collections.Counter(temp_stations[station])[0]
            except KeyboardInterrupt:
                nr_of_zeros = 0

            # --
            # Count number of zeros which follow
            # --
            max_cnt_zeros = count_sequence_of_zeros(temp_stations[station])

            if nr_of_nans > crit_missing_values or nr_of_zeros > crit_nr_of_zeros or max_cnt_zeros > nr_daily_zeros:
                print("Zeros in sequence: {} nr_of_nans: {} nr_of_zeros: {} Ignored station: {} {}".format(
                    max_cnt_zeros,
                    nr_of_nans,
                    nr_of_zeros,
                    station,
                    year), flush=True)
            else:
                # Interpolate missing np.nan values
                temp_stations[station][nans] = np.interp(
                    x(nans),
                    x(~nans),
                    temp_stations[station][~nans])

                # test if still np.nan value
                list_with_all_nan_args = list(np.argwhere(np.isnan(temp_stations[station])))

                if list_with_all_nan_args:
                    raise Exception("Still has np.nan entries")

                # Replace with day, hour array
                interpolated_values_reshaped = temp_stations[station].reshape(365, 24)
                temp_stations_cleaned_reshaped[station] = interpolated_values_reshaped

        # Write temperature data out to csv file
        for station_name, temp_values in temp_stations_cleaned_reshaped.items():

            file_name = "{}__{}.{}".format(year, station_name, "txt")

            path_out = os.path.join(path_out_files, file_name)

            np.savetxt(
                path_out,
                temp_values,
                delimiter=",")
        print("--Number of stations '{}'".format(len(list(temp_stations_cleaned_reshaped.keys()))))

    print("... finished cleaning weather data")

run(
    path_files="//linux-filestore.ouce.ox.ac.uk/mistral/nismod/data/energy_demand/H-Met_office_weather_data/_complete_meteo_data_all_yrs",
    path_out_files="//linux-filestore.ouce.ox.ac.uk/mistral/nismod/data/energy_demand/H-Met_office_weather_data/_complete_meteo_data_all_yrs_cleaned",
    crit_missing_values=40,
    crit_nr_of_zeros=500,
    nr_daily_zeros=20)


"""Functions related to load profiles
"""
import numpy as np

class LoadProfileStock(object):
    """Collection of load shapes in a list

    Arguments
    ----------
    name : string
        Load profile stock name
    """
    def __init__(self, name):
        self.name = name
        self.load_profiles = {}
        self.dict_tuple_keys = {}
        self.stock_enduses = set([])

    def add_lp(
            self,
            unique_identifier,
            technologies,
            enduses,
            shape_yd,
            shape_y_dh,
            model_yeardays,
            sectors,
            shape_yh=False
        ):
        """Add load profile to stock

        Arguments
        ---------
        unique_identifier : str
            Name (unique identifier)
        technologies : list
            Technologies for which the profile applies
        enduses : list
            Enduses for which the profile applies
        shape_y_dh : array

        shape_yh : array
            Shape yh (from year to hour)
        sectors : list, default=False
            Sectors for which the profile applies
        """
        self.load_profiles[unique_identifier] = LoadProfile(
            enduses,
            unique_identifier,
            shape_yh,
            shape_yd,
            shape_y_dh,
            model_yeardays)

        # Generate lookup dictionary with triple key
        self.dict_tuple_keys = generate_key_lu_dict(
            self.dict_tuple_keys,
            unique_identifier,
            enduses,
            sectors,
            technologies)

        # Update enduses in stock
        self.stock_enduses = get_stock_enduses(self.load_profiles)

    def get_lp(self, enduse, sector, technology, shape):
        """Get shape for a certain technology, enduse and sector

        Arguments
        ----------
        enduse : str
            Enduse
        sector : str
            Sector
        technology : str
            technology
        shape : str
            Type of shape which is to be read out from 'load_profiles'

        Return
        ------
        Load profile attribute
        """
        # Get key from lookup dict
        position_in_dict = self.dict_tuple_keys[(enduse, sector, technology)]

        # Get correct object
        load_profile_obj = self.load_profiles[position_in_dict]

        return getattr(load_profile_obj, shape)

def generate_key_lu_dict(
        dict_tuple_keys,
        unique_identifier,
        enduses,
        sectors,
        technologies
    ):
    """Generate look_up keys to position in 'load_profiles'

    Arguments
    ----------
    dict_tuple_keys : dict
        Already existing lu keys
    unique_identifier : string
        Unique identifier of load shape object
    enduses : list
        List with enduses
    sectors : list
        List with sectors
    technologies : list
        List with technologies

    Returns
    -------
    dict_tuple_keys : str
        Lookup position in dict
    """
    for enduse in enduses:
        for sector in sectors:
            for technology in technologies:
                dict_tuple_keys[(enduse, sector, technology)] = unique_identifier

    return dict_tuple_keys

def get_stock_enduses(load_profiles):
    """Update the list of the object with all
    enduses for which load profies are provided

    Arguments
    ---------
    load_profiles : dict
        All load profiles of load profile stock

    Returns
    ------
    all_enduses : list
        All enduses in stock
    """
    all_enduses = set([])
    for profile_obj in load_profiles.values():
        for enduse in profile_obj.enduses:
            all_enduses.add(enduse)

    return list(all_enduses)

class LoadProfile(object):
    """Load profile container to store differengt shapes

    Arguments
    ----------
    enduses : list
        Enduses assigned to load profile
    unique_identifier : string
        Unique identifer for LoadProfile object
    shape_y_dh : array
        Shape of every day in a year (sum = 365)
    shape_yh : array
        Shape yh (from year to hour)
        Standard value is average daily amount
    shape_peak_dh : array
        Shape (dh), shape of a day for every hour
    """
    def __init__(
            self,
            enduses,
            unique_identifier,
            shape_yh,
            shape_yd,
            shape_y_dh,
            model_yeardays
        ):
        """Constructor
        """
        self.unique_identifier = unique_identifier
        self.enduses = enduses

        if isinstance(shape_yh, bool):
            self.shape_yh = calc_yh(
                shape_yd,
                shape_y_dh,
                model_yeardays)
        else:
            self.shape_yh = shape_yh

def abs_to_rel(absolute_array):
    """Convert absolute numbers in an array to relative

    Arguments
    ----------
    absolute_array : array
        Contains absolute numbers in it

    Returns
    -------
    relative_array : array
        Contains relative numbers

    Note
    ----
    - If the total sum is zero, return an array with zeros
    """
    sum_array = float(np.sum(absolute_array))
    if sum_array != 0.0:
        relative_array = absolute_array / sum_array
        relative_array[np.isnan(relative_array)] = 0
        return relative_array
    else:
        return absolute_array

def calk_peak_h_dh(fuel_peak_dh):
    """Ger peak hour in peak day

    Arguments
    ----------
    fuel_peak_dh : array
        Fuel of peak day for every fueltype

    Return
    ------
    peak_fueltype_h : array
        Fuel for maximum hour in peak day per fueltype
    """
    # Get maximum value per row (maximum fuel hour per fueltype)
    peak_fueltype_h = np.max(fuel_peak_dh, axis=1)

    return peak_fueltype_h

def calk_peak_h_dh_single_fueltype(fuel_peak_dh):
    """Ger peak hour in peak day

    Arguments
    ----------
    fuel_peak_dh : array
        Fuel of peak day for every fueltype

    Return
    ------
    peak_fueltype_h : array
        Fuel for maximum hour in peak day per fueltype
    """
    # Get maximum value per row (maximum fuel hour per fueltype)
    peak_fueltype_h = np.max(fuel_peak_dh, axis=0)

    return peak_fueltype_h

def calc_av_lp(demand_yh, seasons, model_yeardays_daytype):
    """Calculate average load profile for daytype and season
    for fuel of a fueltype

    Result
    ------
    demand_yh : array
        Energy demand for every day of a single fueltype
    seasons: dict
        Seasons and their yeardays
    model_yeardays_daytype : dict
        Yearday type of every year
    av_loadprofiles : dict
        season, daytype

    Returns
    -------
    av_season_daytypes : dict
        Averaged lp
    season_daytypes : dict
        Not averaged lp

    """
    season_daytypes = {
        'spring': {
            'workday': np.zeros((0, 24), dtype="float"),
            'holiday': np.zeros((0, 24), dtype="float")},
        'summer': {
            'workday': np.zeros((0, 24), dtype="float"),
            'holiday': np.zeros((0, 24), dtype="float")},
        'autumn': {
            'workday': np.zeros((0, 24), dtype="float"),
            'holiday': np.zeros((0, 24), dtype="float")},
        'winter': {
            'workday': np.zeros((0, 24), dtype="float"),
            'holiday': np.zeros((0, 24), dtype="float")}}

    av_season_daytypes = {
        'spring': {
            'workday': np.zeros((0, 24), dtype="float"),
            'holiday': np.zeros((0, 24), dtype="float")},
        'summer': {
            'workday': np.zeros((0, 24), dtype="float"),
            'holiday': np.zeros((0, 24), dtype="float")},
        'autumn': {
            'workday': np.zeros((0, 24), dtype="float"),
            'holiday': np.zeros((0, 24), dtype="float")},
        'winter': {
            'workday': np.zeros((0, 24), dtype="float"),
            'holiday': np.zeros((0, 24), dtype="float")}}

    for yearday, daytype_yearday in enumerate(model_yeardays_daytype):

        # Season
        if yearday in seasons['spring']:
            season = 'spring'
        elif yearday in seasons['summer']:
            season = 'summer'
        elif yearday in seasons['autumn']:
            season = 'autumn'
        else:
            season = 'winter'

        # Add data as row to array
        new_data_dh = demand_yh[yearday]
        existing_array = season_daytypes[season][daytype_yearday]

        # Add to dict
        season_daytypes[season][daytype_yearday] = np.vstack([existing_array, new_data_dh])

    # -----------------------------
    # Calculate average of all dict
    # -----------------------------
    # Calculate average over every hour in a day
    for season, daytypes_data in season_daytypes.items():
        for daytype, daytpe_data in daytypes_data.items():
            av_season_daytypes[season][daytype] = np.average(daytpe_data, axis=0)

    return av_season_daytypes, season_daytypes

def calc_yh(shape_yd, shape_y_dh, model_yeardays):
    """Calculate the shape based on yh and y_dh shape

    Arguments
    ---------
    shape_yd : array
        Shape with fuel amount for every day (365)
    shape_y_dh : array
        Shape for every day (365, 24), total sum = 365
    model_yeardays : array
        Modelled yeardays

    Returns
    -------
    shape_yh : array
        Shape for every hour in a year (total sum == 1)
    """
    shape_yh = shape_yd[:, np.newaxis] * shape_y_dh[model_yeardays]

    return shape_yh



"""Plot local and regional figure
"""
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from scipy import stats

from energy_demand.technologies import tech_related
from energy_demand.plotting import basic_plot_functions

def run_fig_p2_temporal_validation(
        data_input,
        weather_yr,
        fueltype_str,
        simulation_yr_to_plot,
        period_h,
        validation_elec_2015,
        non_regional_elec_2015,
        fig_name,
        titel="titel",
        y_lim_val=100,
        plot_validation=False,
        plot_show=False
    ):
    """
    """
    fueltype_int = tech_related.get_fueltype_int(fueltype_str)

    # -----------------------------------------------------------
    # Iterate overall weather_yrs and store data in dataframe
    # (columns = timestep, rows: value of year)
    # -----------------------------------------------------------
    selection_hours = period_h

    # List of selected data for every weather year (which is then converted to array)
    weather_yrs_data = []
    weather_yrs_full_year = []

    for station_id, data_weather_yr in data_input[weather_yr].items():

        # Weather year specific data
        data_input_fueltype = data_weather_yr[simulation_yr_to_plot][fueltype_int]     # Select fueltype
        data_input_reshape = data_input_fueltype.reshape(8760)  # reshape
        data_input_selection_hrs = data_input_reshape[selection_hours] # select period

        weather_yrs_data.append(data_input_selection_hrs)
        weather_yrs_full_year.append(data_input_reshape)

    weather_yrs_data = np.array(weather_yrs_data)

    # Create dataframe
    df = pd.DataFrame(weather_yrs_data, columns=period_h)

    df_full_year = pd.DataFrame(weather_yrs_full_year, columns=range(8760))

    # Calculate quantiles
    quantile_95 = 0.95
    quantile_05 = 0.05

    df_q_95 = df.quantile(quantile_95)
    df_q_05 = df.quantile(quantile_05)

    #Transpose for plotting purposes
    df = df.T
    df_q_95 = df_q_95.T
    df_q_05 = df_q_05.T

    fig = plt.figure()

    ax = fig.add_subplot(111)

    # ---------------
    # Smoothing lines
    # ---------------
    try:
        period_h_smoothed, df_q_95_smoothed = basic_plot_functions.smooth_data(period_h, df_q_95, num=40000)
        period_h_smoothed, df_q_05_smoothed = basic_plot_functions.smooth_data(period_h, df_q_05, num=40000)
        #period_h_smoothed, data_2015_smoothed = basic_plot_functions.smooth_data(period_h, data_2015, num=40000)
    except:
        period_h_smoothed = period_h
        df_q_95_smoothed = df_q_95
        df_q_05_smoothed = df_q_05

    #plt.plot(period_h_smoothed, data_2015_smoothed, color='tomato', linestyle='-', linewidth=2, label="2015 weather_yr")
    plt.plot(period_h_smoothed, df_q_05_smoothed, color='black', linestyle='--', linewidth=0.5, label="0.05")
    plt.plot(period_h_smoothed, df_q_95_smoothed, color='black', linestyle='--', linewidth=0.5, label="0.95")

    # -----------------
    # Uncertainty range
    # -----------------
    plt.fill_between(
        period_h_smoothed, #x
        df_q_95_smoothed,  #y1
        df_q_05_smoothed,  #y2
        alpha=1,
        facecolor="lightgrey",
        label="uncertainty band")

    # -----------------
    # All weather stations are used for this data
    # -----------------
    all_weather_stations_2015 = non_regional_elec_2015.reshape(8760)[period_h] 

    # -----------------
    # Validation data
    # -----------------
    if plot_validation:
        validation_2015 = validation_elec_2015.reshape(8760)[selection_hours] 

        try:
            period_h_smoothed, validation_2015_smoothed = basic_plot_functions.smooth_data(period_h, validation_2015, num=40000)
        except:
            period_h_smoothed = period_h
            validation_2015_smoothed = validation_2015

        plt.plot(
            period_h_smoothed,
            validation_2015_smoothed,
            color='green',
            linestyle='--',
            linewidth=1.5, label="validation 2015")

        # -----------
        # statistics
        # -----------
        # Calculate mean of all all single station runs
        mean_all_single_runs = df_full_year.mean(axis=0).tolist()
        mean_all_single_runs = np.array(mean_all_single_runs)[period_h]

        slope, intercept, r_value, p_value, std_err = stats.linregress(
            validation_2015,
            all_weather_stations_2015)
        slope, intercept, r_value2, p_value, std_err = stats.linregress(
            validation_2015,
            mean_all_single_runs)

        print("R_Value_all_stations: " + str(r_value))
        print("R_Value_single_stations: " + str(r_value2))
        plt.title(
            "R_value: all stations: {} mean_single_weather_stations: {}".format(
                round(r_value, 2),
                round(r_value2, 2)))
    
    try:
        period_h_smoothed, all_weather_stations_2015_smoothed = basic_plot_functions.smooth_data(
            period_h, all_weather_stations_2015, num=40000)
    except:
        period_h_smoothed = period_h
        all_weather_stations_2015_smoothed = all_weather_stations_2015

    plt.plot(
        period_h_smoothed,
        all_weather_stations_2015_smoothed,
        color='blue',
        linestyle='--',
        linewidth=1.2,
        label="all stations")

    # Ticks for single day
    #major_ticks_days, major_ticks_labels = get_date_strings(
    #    days_to_plot,
    #    daystep=1)
    #plt.xticks(major_ticks_days, major_ticks_labels)
    zero_entry = period_h[0]
    plt.xticks([0 + zero_entry, 5 + zero_entry, 11 + zero_entry, 17 + zero_entry, 23 + zero_entry], ['1', '6', '12', '18', '24'])

    plt.ylim(0, y_lim_val)
    plt.xlim(period_h[0], period_h[-1])
    plt.title("Peak day: " + str(titel))
    plt.legend(
        prop={
            'family':'arial',
            'size': 10},
        loc='best',
        frameon=False,
        shadow=True)

    if plot_show:
        plt.show()
    else:
        pass
    plt.savefig(fig_name)
    plt.close()



"""Download population projections from https://github.com/nismod/population/blob/master/README.md


Info
-----

https://github.com/virgesmith/UKCensusAPI
https://www.nomisweb.co.uk/myaccount/webservice.asp
https://github.com/nismod/population
https://github.com/virgesmith/UKCensusAPI

Steps
------
1. optain nomis key
2. in cmd line: set NOMIS_API_KEY=XXX
3. ./setup.py install
4. run python script from command line

https://www.ons.gov.uk/peoplepopulationandcommunity/populationandmigration/populationprojections/bulletins/nationalpopulationprojections/2015-10-29

Potential variants
-------------------
hhh: High population,
hpp: High fertility,
lll: Low population,
lpp: Low fertility,
php: High life expectancy,
pjp: Moderately high life expectancy,
pkp: Moderately low life expectancy,
plp: Low life expectancy,
pph: High migration,
ppl: Low migration,
ppp: Principal,
ppq: 0% future EU migration (non-ONS),
ppr: 50% future EU migration (non-ONS),
pps: 150% future EU migration (non-ONS),
ppz: Zero net migration

Select for paper I
-------------------
Prinicpal projection    ppp
Low mnigration          ppl
High migration          pph
"""
import pandas as pd
import population.nppdata as NPPData
import population.snppdata as SNPPData
import population.utils as utils


extrapolate_specific_scenario = True   # 2016-2050
base_year_data = True                  # 2015

if base_year_data:
    """Base year data (no variants, same for all variants)
    """
    npp = NPPData.NPPData()
    snpp = SNPPData.SNPPData()

    year = 2015

    # start with an empty data frame
    result_ppp = pd.DataFrame()

    # loop over all the UK LAD (or LAD-equivalents)
    for lad in snpp.data.GEOGRAPHY_CODE.unique():

        #region_ppp = snpp.create_variant('ppp', npp, lad, year)
        region_ppp = snpp.extrapolagg(["GENDER", "C_AGE"], npp, lad, year) #princiapl

        # aggregate the calculated variants by age and gender
        result_ppp = result_ppp.append(region_ppp, ignore_index=True)

    # write out results
    result_ppp.to_csv("C:/Users/cenv0553/mistral_population/__RESULTS/all_variants_2015.csv", index=False)
    print("Finished writting 2015 data")

if extrapolate_specific_scenario:
    '''
    Get extrapolated data for full time range for different ONS scenarios from 2016 - 2050
    # https://github.com/nismod/population/blob/master/doc/example_variant_ex.py
    '''

    # initialise the population modules
    npp = NPPData.NPPData()
    snpp = SNPPData.SNPPData()

    # 50 years, roughly half is extrapolated
    # Must start with 2015.
    years = range(2016, 2051)

    # start with an empty data frame
    result_ppp = pd.DataFrame()
    result_ppl = pd.DataFrame()
    result_pph = pd.DataFrame()

    # loop over all the UK LAD (or LAD-equivalents)
    for lad in snpp.data.GEOGRAPHY_CODE.unique():
        print("Collection population for region {}".format(lad))

        # Principle variant
        region_ppp = snpp.extrapolagg(["GENDER", "C_AGE"], npp, lad, years)
            
        # calculate the the variants
        #region_ppp = snpp.create_variant("ppp", npp, lad, years)
        region_ppl = snpp.create_variant("ppl", npp, lad, years)
        region_pph = snpp.create_variant("pph", npp, lad, years)

        # aggregate the calculated variants by age and gender
        region_ppl = utils.aggregate(region_ppl, ["GENDER", "C_AGE"])
        region_pph = utils.aggregate(region_pph, ["GENDER", "C_AGE"])

        result_ppp = result_ppp.append(region_ppp, ignore_index=True)
        result_ppl = result_ppl.append(region_ppl, ignore_index=True)
        result_pph = result_pph.append(region_pph, ignore_index=True)

    # write out results
    result_ppp.to_csv("C:/Users/cenv0553/mistral_population/__RESULTS/prinicpal_extrap_2016-2050.csv", index=False)
    result_ppl.to_csv("C:/Users/cenv0553/mistral_population/__RESULTS/lowmigration_extrap_2016-2050.csv", index=False)
    result_pph.to_csv("C:/Users/cenv0553/mistral_population/__RESULTS/highmigration_extrap_2016-2050.csv", index=False)
    print("Finished writting 2016 - 2050 data")



import os
import subprocess
from multiprocessing import Pool, cpu_count

def my_function(simulation_number):  
    print('simulation_number ' + str(simulation_number))

    run_name = 'h_max'
    name_config_path = run_name

    run_smif = False 

    # --------------------------
    # Weather realisation
    # --------------------------
    all_weather_reations = ["NF{}".format(i) for i in range(1, 101, 1)]
    weather_realisation = all_weather_reations[simulation_number]

    path_to_ini_file = "/soge-home/staff/cenv0553/energy_demand/local_run_config_file_cluster.ini"

    # Make run name_specifiv
    run_name = "{}_{}".format(run_name, simulation_number)

    # Run energy demand main.py
    if run_smif:
        # Run smif
        #bash_command = "smif -v run ed_constrained_pop-baseline16_econ-c16_fuel-c16"
        #os.system(bash_command)
        pass
    else:
        bash_command = "python energy_demand/energy_demand/main.py {} {} {} {}".format(
            path_to_ini_file,
            run_name,
            weather_realisation,
            name_config_path)

        os.system(bash_command)
    return 

# Simulation number
simulation_number = range(0, 101, 1)

if __name__ == "__main__":
    with Pool(int(cpu_count()/2)) as pool:
        pool.map(
            my_function,
            simulation_number,
            chunksize=1)



"""Strategy variable assumptions provided as parameters to smif
"""
from collections import defaultdict

from energy_demand.read_write import narrative_related

def generate_default_parameter_narratives(
        default_streategy_vars,
        end_yr=2050,
        base_yr=2015):
    """Load default parameters and create default timesteps

    Arguments
    ---------
    default_streategy_vars : dict
        Default parameter values
    end_yr : int
        Simulation end year
    base_yr : int
        Base year

    Returns
    -------

    """
    strategy_vars = defaultdict(dict)
    # ------------------------------------------------------------
    # Create default narrative for every simulation parameter
    # ------------------------------------------------------------
    for var_name, var_entries in default_streategy_vars.items():
        crit_single_dim = narrative_related.crit_dim_var(var_entries)

        if crit_single_dim:
            scenario_value = var_entries['default_value']

            # Create default narrative with only one timestep from simulation base year to simulation end year
            strategy_vars[var_name] = narrative_related.default_narrative(
                end_yr=end_yr,
                value_by=var_entries['default_value'],                # Base year value,
                value_ey=scenario_value,
                diffusion_choice=var_entries['diffusion_type'],       # Sigmoid or linear,
                base_yr=base_yr,
                regional_specific=var_entries['regional_specific'])   # Criteria whether the same for all regions or not

        else:
            # Standard narrative for multidimensional narrative
            for sub_var_name, sub_var_entries in var_entries.items():

                try:
                    scenario_value = sub_var_entries['scenario_value']
                except TypeError:
                    raise TypeError("{}: Var_entries dict: {}".format(var_name, var_entries))
                # -----------------------------------
                # Crate single-step default narratives (up to end_year)
                # -----------------------------------
                strategy_vars[var_name][sub_var_name] = narrative_related.default_narrative(
                    end_yr=end_yr,
                    value_by=sub_var_entries['default_value'],
                    value_ey=scenario_value,
                    diffusion_choice=sub_var_entries['diffusion_type'],
                    base_yr=base_yr,
                    regional_specific=sub_var_entries['regional_specific'])

    strategy_vars = dict(strategy_vars)

    return strategy_vars

def load_param_assump(
        default_values=None,
        hard_coded_default_val=True
    ):
    """All assumptions of the energy demand model
    are loaded and added to the data dictionary

    Arguments
    ---------

    Returns
    -------
    data : dict
        Data dictionary with added ssumption dict
    """
    strategy_vars = defaultdict(dict)

    if hard_coded_default_val:
        default_values = {
            'ss_t_cooling': 5,
            'is_t_heating': 15.5,
            'gshp_fraction': 0.1,
            'p_cold_rolling_steel_by': 0.2,
            'rs_t_heating': 15.5,
            'smart_meter_p_by': 0.05,
            'cooled_ss_floorarea_by': 0.35,
            'speed_con_max': 1,
            'spatial_explicit_diffusion': 0,
            'ss_t_heating': 15.5}

    default_enduses = {

        # Submodel Residential
        'rs_space_heating': 0,
        'rs_water_heating': 0,
        'rs_lighting': 0,
        'rs_cooking': 0,
        'rs_cold': 0,
        'rs_wet': 0,
        'rs_consumer_electronics': 0,
        'rs_home_computing': 0,

        # Submodel Service (Table 5.5a)
        # same % improvements from baseline over all sectors
        'ss_space_heating': 0,
        'ss_water_heating': 0,
        'ss_cooling_humidification': 0,
        'ss_fans': 0,
        'ss_lighting': 0,
        'ss_catering': 0,
        'ss_small_power': 0,
        'ss_ICT_equipment': 0,
        'ss_cooled_storage': 0,
        'ss_other_gas': 0,
        'ss_other_electricity': 0,

        # Submodel Industry
        # same % improvements from baseline over all sectors
        'is_high_temp_process': 0,
        'is_low_temp_process': 0,
        'is_drying_separation': 0,
        'is_motors': 0,
        'is_compressed_air': 0,
        'is_lighting': 0,
        'is_space_heating': 0,
        'is_other': 0,
        'is_refrigeration': 0}

    # ------------------
    # Spatial explicit diffusion
    # ------------------
    strategy_vars['spatial_explicit_diffusion'] = {
        "name": "spatial_explicit_diffusion",
        "description": "Criteria to define spatial or non spatial diffusion",
        "default_value": default_values['spatial_explicit_diffusion'],
        "sector": True,
        'regional_specific': False,
        'diffusion_type': 'linear'}

    strategy_vars['speed_con_max'] = {
        "name": "speed_con_max",
        "description": "Maximum speed of penetration (for spatial explicit diffusion)",
        "default_value": default_values['speed_con_max'],
        "sector": True,
        'regional_specific': False,
        'diffusion_type': 'linear'}

    # ----------------------
    # Heat pump technology mix
    # Source: Hannon 2015: Raising the temperature of the UK heat pump market: Learning lessons from Finland
    # ----------------------
    strategy_vars['gshp_fraction'] = {
        "name": "gshp_fraction",
        "description": "Relative GSHP (%) to GSHP+ASHP",
        "default_value": default_values['gshp_fraction'],
        "sector": True,
        'regional_specific': False,
        'diffusion_type': 'linear'}

    # ============================================================
    #  Demand management assumptions (daily demand shape)
    #  An improvement in load factor improvement can be assigned
    #  for every enduse (peak shaving)
    #
    #  Example: 0.2 --> Improvement in load factor until ey
    # ============================================================

    # Helper function to create description of parameters for all enduses
    for demand_name, scenario_value in default_enduses.items():
        strategy_vars['dm_improvement'][demand_name] = {
            "name": demand_name,

            "description": "reduction in load factor for enduse {}".format(demand_name),
            "default_value": scenario_value,
            "sector": True,
            "enduse": [demand_name],
            'regional_specific': True,
            'diffusion_type': 'linear'}

    # ============================================================
    # Base temperature assumptions for heating and cooling demand
    # The diffusion is asumed to be linear
    # ============================================================
    strategy_vars['rs_t_base_heating'] = {
        "name": "rs_t_base_heating",
        "description": "Base temperature assumption residential heating",
        "default_value": default_values['rs_t_heating'],
        "sector": True,
        'regional_specific': False,
        'diffusion_type': 'linear'}

    # Future base year temperature
    strategy_vars['ss_t_base_heating'] = {
        "name": "ss_t_base_heating",
        "description": "Base temperature assumption service sector heating",
        "default_value": default_values['ss_t_heating'],
        "sector": True,
        'regional_specific': False,
        'diffusion_type': 'linear'}

    # Cooling base temperature
    # Future base year temperature
    strategy_vars['ss_t_base_cooling'] = {
        "name": "ss_t_base_cooling",
        "description": "Base temperature assumption service sector cooling",
        "default_value": default_values['ss_t_cooling'],
        "sector": True,
        'regional_specific': False,
        'diffusion_type': 'linear'}

    # Future base year temperature
    strategy_vars['is_t_base_heating'] = {
        "name": "is_t_base_heating",
        "description": "Base temperature assumption service sector heating",
        "default_value": default_values['is_t_heating'],
        "sector": True,
        'regional_specific': False,
        'diffusion_type': 'linear'}

    # ============================================================
    # Smart meter penetration
    # ============================================================
    strategy_vars['smart_meter_p'] = {
        "name": "smart_meter_p",
        "description": "Improvement of smart meter penetration",
        "default_value": default_values['smart_meter_p_by'],
        "sector": True,
        'regional_specific': True,
        'diffusion_type': 'linear'}

    # ============================================================
    # Cooling
    # ============================================================
    cooled_floorarea = {
        'ss_cooling_humidification': default_values['cooled_ss_floorarea_by'],
        'ss_fans': default_values['cooled_ss_floorarea_by']}

    for sub_param_name, sub_param_value in cooled_floorarea.items():
        strategy_vars['cooled_floorarea'][sub_param_name] = {
            "name": sub_param_name,

            "description": "Change in cooling of floor area (service sector)",
            "default_value": sub_param_value,
            "sector": True,
            'regional_specific': True,
            'diffusion_type': 'linear'}

    # ============================================================
    # Industrial processes
    # ============================================================
    strategy_vars['p_cold_rolling_steel'] = {
        "name": "p_cold_rolling_steel",
        "description": "Sectoral share of cold rolling in steel manufacturing)",
        "default_value": default_values['p_cold_rolling_steel_by'],
        "sector": True,
        'regional_specific': True,
        'diffusion_type': 'linear'}

    # ============================================================
    # Heat recycling & reuse
    # ============================================================
    heat_recovered = {
        'rs_space_heating': 0,
        'ss_space_heating': 0,
        'is_space_heating': 0,
        'rs_water_heating': 0,
        'ss_water_heating': 0}

    for sub_param_name, sub_param_value in heat_recovered.items():
        strategy_vars['heat_recovered'][sub_param_name] = {
            "name": sub_param_name,

            "description": "Reduction in heat because of heat recovery and recycling",
            "default_value": sub_param_value,
            "sector": True,
            "enduse": [sub_param_name],
            'regional_specific': True,
            'diffusion_type': 'linear'}

    # ============================================================
    # Air leakage
    # ============================================================
    air_leakage = {
        'rs_space_heating': 0,
        'ss_space_heating': 0,
        'is_space_heating': 0}

    for sub_param_name, sub_param_value in air_leakage.items():
        strategy_vars['air_leakage'][sub_param_name] = {
            "name": sub_param_name,

            "description": "Reduction in heat because of air leakage improvement (residential sector)",
            "default_value": sub_param_value,
            "sector": True,
            "enduse": [sub_param_name],
            'regional_specific': True,
            'diffusion_type': 'linear'}

    # ---------------------------------------------------------
    # General change in fuel consumption for specific enduses
    # ---------------------------------------------------------
    #   With these assumptions, general efficiency gain (across all fueltypes)
    #   can be defined for specific enduses. This may be e.g. due to general
    #   efficiency gains or anticipated increases in demand.
    #   NTH: Specific hanges per fueltype (not across al fueltesp)
    #
    #   Change in fuel until the simulation end year (
    #   if no change set to 1, if e.g. 10% decrease change to 0.9)
    # -------------------------------------------------------

    # Helper function to create description of parameters for all enduses
    for enduse, param_value in default_enduses.items():
        strategy_vars['generic_enduse_change'][enduse] = {
            "name": enduse,
            "description": "Enduse specific change {}".format(enduse),
            "default_value": param_value,
            "sector": True,
            "enduse": [enduse],
            "sector": [],
            'regional_specific': True,
            'diffusion_type': 'linear'}

    # ============================================================
    # Technologies & efficiencies
    # ============================================================

    # --Assumption how much of technological efficiency is reached
    strategy_vars["f_eff_achieved"] = {
        "name": "f_eff_achieved",
        "description": "Fraction achieved of efficiency improvements",
        "default_value": 0, # Default is no efficiency improvement
        "sector": True,
        'regional_specific': False,
        'diffusion_type': 'linear'}

    # --------------------------------------
    # Floor area per person change
    # ---------------------------------------
    strategy_vars["assump_diff_floorarea_pp"] = {
        "name": "assump_diff_floorarea_pp",
        "description": "Change in floor area per person (%, 1=100%)",
        "default_value": 0,
        "sector": True,
        'regional_specific': False,
        'diffusion_type': 'linear'}

    # -----------------------
    # Generic enduse and sector specific fuel switches
    # -----------------------
    for enduse, param_value in default_enduses.items():
        strategy_vars["generic_fuel_switch"][enduse] = {
            "name": "generic_fuel_switch",
            "description": "Generic fuel switches to switch fuel in any enduse and sector",
            "default_value": param_value,
            "enduse": enduse,
            "sector": True,
            'regional_specific': True,
            'diffusion_type': 'linear'}

    strategy_vars_out = autocomplete_strategy_vars(strategy_vars)

    return dict(strategy_vars_out)

def autocomplete_strategy_vars(
        strategy_vars,
        narrative_crit=False
    ):
    """Autocomplete all narratives or strategy variables with
    and 'enduse' or 'sector' in case they are not defined.

    Arguments
    ----------
    strategy_vars

    Returns
    -------
    narrative_crit : bool
        Criteria wheter inputs are a narrative or not
    """
    if not narrative_crit:
        out_dict = defaultdict(dict)

        for var_name, var_entries in strategy_vars.items():
            crit_single_dim = narrative_related.crit_dim_var(var_entries)

            if crit_single_dim:
                out_dict[var_name] = var_entries

                # If no 'enduse' defined, add empty list of affected enduses
                out_dict[var_name]['scenario_value'] = var_entries['default_value']

                if 'enduse' not in var_entries:
                    out_dict[var_name]['enduse'] = []
                if 'sector' not in var_entries:
                    out_dict[var_name]['sector'] = True  # All sector
            else:
                for sub_var_name, sub_var_entries in var_entries.items():
                    out_dict[var_name][sub_var_name] = sub_var_entries

                    out_dict[var_name][sub_var_name]['scenario_value'] = sub_var_entries['default_value']

                    # If no 'enduse' defined, add empty list of affected enduses
                    if 'enduse' not in sub_var_entries:
                        out_dict[var_name][sub_var_name]['enduse'] = []
                    if 'sector' not in sub_var_entries:
                        out_dict[var_name][sub_var_name]['sector'] = True # All sector
    else:
        # Same but narratives which need to be iterated
        out_dict = {}

        for var_name, var_entries in strategy_vars.items():
            out_dict[var_name] = {}

            crit_single_dim = narrative_related.crit_dim_var(var_entries)

            if crit_single_dim:
                updated_narratives = []

                for narrative in var_entries:
                    #If no 'enduse' defined, add empty list of affected enduses
                    if 'enduse' not in narrative:
                        narrative['enduse'] = []
                    if 'sector' not in narrative:
                        narrative['sector'] = 'dummy_sector' # All sector
                    updated_narratives.append(narrative)

                out_dict[var_name] = updated_narratives
            else:
                #print("   ...user defined variable: %s", var_name)
                for sub_var_name, sector_sub_var_entries in var_entries.items():

                    if type(sector_sub_var_entries) is dict: # If sectors are defined
                        for sector, sub_var_entries in sector_sub_var_entries.items():
                            out_dict[var_name][sub_var_name] = {}

                            updated_narratives = []
                            for narrative in sub_var_entries:

                                # If no 'enduse' defined, add empty list of affected enduses
                                if 'enduse' not in narrative:
                                    narrative['enduse'] = []
                                if 'sector' not in narrative:
                                    narrative['sector'] = sector
                                updated_narratives.append(narrative)

                            out_dict[var_name][sub_var_name][sector] = updated_narratives

                    else: # no sectors defined
                        updated_narratives = []
                        for narrative in sector_sub_var_entries:

                            if 'enduse' not in narrative:
                                narrative['enduse'] = [sub_var_name]
                            updated_narratives.append(narrative)

                        out_dict[var_name][sub_var_name] = updated_narratives

    return dict(out_dict)

def get_affected_enduse(strategy_vars, name):
    """Get all defined affected enduses of a scenario variable

    Arguments
    ---------
    strategy_vars : dict
        Dict with all defined strategy variables
    name : str
        Name of variable to get

    Returns
    -------
    enduses : list
        AFfected enduses of scenario variable
    """
    try:
        for var in strategy_vars:
            if var['name'] == name:
                enduses = var['enduse']

                return enduses
    except KeyError: # Not affected enduses defined
        enduses = []

        return enduses



"""Testing
"""
import numpy as np
from energy_demand.technologies import tech_related
from energy_demand.read_write import read_data

'''def test_get_tech_type():
    """Testing
    """
    tech_list = {
        'heating_non_const': ['heat_p'],
        'heating_const': ['boilerA'],
        'storage_heating_electricity': ['boilerC'],
        'secondary_heating_electricity': []}

    assert tech_related.get_tech_type('placeholder_tech', tech_list) == 'placeholder_tech'
    assert tech_related.get_tech_type('heat_p', tech_list) == 'heat_pump'
    assert tech_related.get_tech_type('test_tech', tech_list) == 'other_tech'''

def test_calc_eff_cy():
    """Testing
    """
    out_value = tech_related.calc_eff_cy(
        base_yr=2015,
        curr_yr=2020,
        eff_by=1.0,
        eff_ey=2.0,
        yr_until_changed=2020,
        f_eff_achieved=1.0,
        diff_method='linear')

    assert out_value == 2.0

    out_value = tech_related.calc_eff_cy(
        base_yr=2015,
        curr_yr=2020,
        eff_by=1.0,
        eff_ey=2.0,
        yr_until_changed=2020,
        f_eff_achieved=1.0,
        diff_method='sigmoid')

    assert out_value == 2.0

def test_calc_hp_eff():
    """Testing function
    """
    temp_yh = np.zeros((365, 24)) + 10

    efficiency_intersect = 10
    t_base_heating = 15.5

    # call function
    out_value = tech_related.calc_hp_eff(
        temp_yh,
        efficiency_intersect,
        t_base_heating)

    float_value = 1.0
    expected = type(float_value)
    assert type(out_value) == expected

def test_eff_heat_pump():
    """Testing
    """
    t_base = 15.5
    temp_yh = np.zeros((365, 24)) + 5 #make diff 10 degrees
    temp_diff = t_base - temp_yh

    efficiency_intersect = 6 #Efficiency of hp at temp diff of 10 degrees

    out_value = tech_related.eff_heat_pump(temp_diff, efficiency_intersect)

    values_every_h = -0.08 * temp_diff + (efficiency_intersect - (-0.8))
    expected = np.average(values_every_h)

    assert out_value == expected

def test_get_fueltype_str():
    """Testing function
    """
    fueltypes = {'gas': 1}
    in_value = 1
    expected = 'gas'

    # call function
    out_value = tech_related.get_fueltype_str(fueltypes, in_value)

    assert out_value == expected

def test_get_fueltype_int():
    """Testing function
    """
    fueltypes = {'gas': 1}
    in_value = 'gas'
    expected = 1

    # call function
    out_value = tech_related.get_fueltype_int(in_value)

    assert out_value == expected

def test_calc_av_heat_pump_eff_ey():
    """testing
    """
    technologies = {
        'heat_pump_ASHP_electricity': read_data.TechnologyData(),
        'heat_pump_GSHP_electricity': read_data.TechnologyData(),
        'heat_pumps_electricity': read_data.TechnologyData()
        }
    technologies['heat_pump_ASHP_electricity'].fueltype_str = 'electricity'
    technologies['heat_pump_ASHP_electricity'].eff_achieved = 1.0
    technologies['heat_pump_ASHP_electricity'].diff_method = 'linear'
    technologies['heat_pump_ASHP_electricity'].eff_by = 0.5
    technologies['heat_pump_ASHP_electricity'].eff_ey = 0.5
    technologies['heat_pump_ASHP_electricity'].year_eff_ey = 2020

    technologies['heat_pump_GSHP_electricity'].fueltype_str = 'electricity'
    technologies['heat_pump_GSHP_electricity'].eff_achieved = 1.0
    technologies['heat_pump_GSHP_electricity'].diff_method = 'linear'
    technologies['heat_pump_GSHP_electricity'].eff_by = 1.0
    technologies['heat_pump_GSHP_electricity'].eff_ey = 1.0
    technologies['heat_pump_GSHP_electricity'].year_eff_ey = 2020

    technologies['heat_pumps_electricity'].fueltype_str = 'electricity'
    technologies['heat_pumps_electricity'].eff_achieved = 1.0
    technologies['heat_pumps_electricity'].diff_method = 'linear'
    technologies['heat_pumps_electricity'].eff_by = 1.0
    technologies['heat_pumps_electricity'].eff_ey = 99999
    technologies['heat_pumps_electricity'].year_eff_ey = 2020

    heat_pump_assump = {
        'electricity': {
            'heat_pump_ASHP_electricity': 0.7,
            'heat_pump_GSHP_electricity': 0.3
            }
    }

    result = tech_related.calc_av_heat_pump_eff_ey(
        technologies,
        heat_pump_assump)

    assert result['heat_pumps_electricity'].eff_ey == 0.7 * 0.5 + 0.3 * 1.0



"""Testing
"""
import numpy as np
from energy_demand.scripts import s_disaggregation
from energy_demand.assumptions import general_assumptions

def test_rs_disaggregate():
    """testing
    """
    dummy_sector = None
    regions = ['regA', 'regB']

    national_fuel = 100
    rs_national_fuel = {'rs_space_heating': {dummy_sector: national_fuel}}

    scenario_data = {
        'population': {2015: {'regA': 10, 'regB': 10}},
        'floor_area': {'rs_floorarea': {2015: {'regA': 10, 'regB': 10}}}}

    assumptions = {
        'base_yr': 2015,
        'curr_yr': 2020,
        'rs_regions_without_floorarea': [],
        'base_temp_diff_params': {
            'sig_midpoint': 0,
            'sig_steepness': 1,
            'yr_until_changed': 2020},
        'strategy_vars': {'rs_t_base_heating': {'scenario_value': 15}}}
    assumptions = general_assumptions.DummyClass(assumptions)
    assumptions.__setattr__('t_bases', general_assumptions.DummyClass({'rs_t_heating': 15}))

    reg_coord = {
        'regA': {'longitude': 0, 'latitude': 0},
        'regB': {'longitude': 0, 'latitude': 0}}

    weather_stations = {
        'stationID_1': {'longitude': 1, 'latitude': 1}}

    temp_data = {'stationID_1': np.ones((365, 24)) + 10}
    enduses = ['rs_space_heating']

    pop_for_disag = {
        2015: {
            'regA': 100,
            'regB': 100},
        2020: {
            'regA': 100,
            'regB': 100}} 

    result = s_disaggregation.rs_disaggregate(
        regions=regions,
        rs_national_fuel=rs_national_fuel,
        scenario_data=scenario_data,
        pop_for_disagg=pop_for_disag,
        assumptions=assumptions,
        reg_coord=reg_coord,
        weather_stations=weather_stations,
        temp_data=temp_data,
        enduses=enduses,
        crit_limited_disagg_pop_hdd=True,
        crit_limited_disagg_pop=False,
        crit_full_disagg=False,
        dummy_sector=dummy_sector)
    
    assert result['regA']['rs_space_heating'][None] == national_fuel / 2

    result = s_disaggregation.rs_disaggregate(
        regions=regions,
        rs_national_fuel=rs_national_fuel,
        scenario_data=scenario_data,
        pop_for_disagg=pop_for_disag,
        assumptions=assumptions,
        reg_coord=reg_coord,
        weather_stations=weather_stations,
        temp_data=temp_data,
        enduses=enduses,
        crit_limited_disagg_pop_hdd=False,
        crit_limited_disagg_pop=True,
        crit_full_disagg=False,
        dummy_sector=dummy_sector)

    assert result['regA']['rs_space_heating'][None] == national_fuel / 2

def test_ss_disaggregate():
    """testing
    """
    regions = ['regA', 'regB']
    national_fuel = 100
    raw_fuel_sectors_enduses = {'ss_space_heating': {'offices': national_fuel}}

    scenario_data = {
        'population': {2015: {'regA': 10, 'regB': 10}},
        'floor_area': {
            'ss_floorarea_newcastle': {
                2015: {'regA': 10, 'regB': 10}},
            'ss_floorarea' :{
                 2015: {'regA': {'offices': 100}, 'regB': {'offices': 100}}}
            },
        }

    assumptions = {
        'base_yr': 2015,
        'curr_yr': 2020,
        'ss_regions_without_floorarea': [],
        'base_temp_diff_params': {
            'sig_midpoint': 0,
            'sig_steepness': 1,
            'yr_until_changed': 2020},
        'strategy_vars': {
            'ss_t_base_heating': {'scenario_value': 15},
            'ss_t_base_cooling': {'scenario_value': 20}}}

    assumptions = general_assumptions.DummyClass(assumptions)
    assumptions.__setattr__('t_bases', general_assumptions.DummyClass(
        {'ss_t_heating': 15, 'ss_t_cooling': 20}))

    reg_coord = {
        'regA': {'longitude': 0, 'latitude': 0},
        'regB': {'longitude': 0, 'latitude': 0}}

    weather_stations = {
        'stationID_1': {'longitude': 1,'latitude': 1}}

    temp_data = {'stationID_1': np.ones((365, 24)) + 10}

    enduses = ['ss_space_heating']
    sectors = ['offices']

    service_building_count = {}
    service_building_count[9] = {}
    service_building_count[9]['regA'] = 10
    service_building_count[9]['regB'] = 10

    pop_for_disag = {
        2015: {
            'regA': 100,
            'regB': 100},
        2020: {
            'regA': 100,
            'regB': 100}}            

    result = s_disaggregation.ss_disaggregate(
        ss_national_fuel=raw_fuel_sectors_enduses,
        service_building_count=service_building_count,
        assumptions=assumptions,
        scenario_data=scenario_data,
        pop_for_disagg=pop_for_disag,
        regions=regions,
        reg_coord=reg_coord,
        temp_data=temp_data,
        weather_stations=weather_stations,
        enduses=enduses,
        sectors=sectors,
        crit_limited_disagg_pop_hdd=False,
        crit_limited_disagg_pop=True,
        crit_full_disagg=False)

    assert result['regA']['ss_space_heating']['offices'] == national_fuel / 2

def test_is_disaggregate():
    """TESTING"""
    temp_data = {'stationID_1': np.ones((365, 24)) + 10}
    reg_coord = {
        'regA': {'longitude': 0, 'latitude': 0},
        'regB': {'longitude': 0, 'latitude': 0}}

    weather_stations = {
        'stationID_1': {'longitude': 1,'latitude': 1}}

    national_fuel = 100
    raw_fuel_sectors_enduses = {'is_space_heating': {
        'mining': national_fuel,
        'pharmaceuticals': national_fuel}}

    assumptions = {
        'base_yr': 2015,
        'curr_yr': 2020,
        'base_temp_diff_params': {
            'sig_midpoint': 0,
            'sig_steepness': 1,
            'yr_until_changed': 2020},
        'strategy_vars': {
            'is_t_heating': {'scenario_value': 0},
            'is_t_base_heating': {'scenario_value': 15},
            'is_t_base_cooling_future_yr': {'scenario_value': 15}}}

    assumptions = general_assumptions.DummyClass(assumptions)
    assumptions.__setattr__('t_bases', general_assumptions.DummyClass({'is_t_heating': 15, 'is_t_cooling': 15}))

    regions = ['regA', 'regB']
    enduses = ['is_space_heating']
    sectors = ['mining', 'pharmaceuticals']

    employment_statistics = {
        'regA': {'B': 0, 'M': 10}, #mining, pharmaceuticals
        'regB': {'B': 5, 'M': 5}}

    pop_for_disag = {
        2015: {
            'regA': 10,
            'regB': 10}}

    result = s_disaggregation.is_disaggregate(
        assumptions=assumptions,
        temp_data=temp_data,
        reg_coord=reg_coord,
        weather_stations=weather_stations,
        is_national_fuel=raw_fuel_sectors_enduses,
        regions=regions,
        enduses=enduses,
        sectors=sectors,
        employment_statistics=employment_statistics,
        pop_for_disagg=pop_for_disag,
        census_disagg=False,
        )

    assert result['regA']['is_space_heating']['mining'] == 50.0

    # ----
    result = s_disaggregation.is_disaggregate(
        assumptions=assumptions,
        temp_data=temp_data,
        reg_coord=reg_coord,
        weather_stations=weather_stations,
        is_national_fuel=raw_fuel_sectors_enduses,
        regions=regions,
        enduses=enduses,
        sectors=sectors,
        employment_statistics=employment_statistics,
        pop_for_disagg=pop_for_disag,
        census_disagg=True)

    assert result['regA']['is_space_heating']['mining'] == 0
    assert round(result['regA']['is_space_heating']['pharmaceuticals'], 3) == round(10.0/15.0 * 100,3) 
    assert result['regB']['is_space_heating']['mining'] == 100
    assert round(result['regB']['is_space_heating']['pharmaceuticals'], 3)  == round(5.0/15.0 * 100, 3)



"""Model initiallisation scripts
"""
import logging
from collections import defaultdict
import numpy as np

from energy_demand.basic import lookup_tables
from energy_demand.geography import spatial_diffusion
from energy_demand.read_write import read_data
from energy_demand.scripts import (s_fuel_to_service, s_generate_sigmoid)
from energy_demand.technologies import fuel_service_switch
from energy_demand.scripts import s_scenario_param
from energy_demand.read_write import narrative_related

def get_all_narrative_timesteps(switches_list):
    """Read all defined narrative timesteps
    from switches

    Arguments
    ---------
    switches : list
        All different switches

    Returns
    -------
    narrative_timesteps : dict
        All defined timesteps in narrative per enduse
        provided as lists in dict
    """
    narrative_timesteps = {}

    enduses = fuel_service_switch.get_all_enduses_of_switches(
        switches_list)

    for enduse in enduses:
        narrative_timesteps[enduse] = set()

        for switch in switches_list:
            if switch.enduse == enduse:
                narrative_timesteps[enduse].add(switch.switch_yr)

        narrative_timesteps[enduse] = list(narrative_timesteps[enduse])
        narrative_timesteps[enduse].sort()

    return narrative_timesteps

def create_spatial_diffusion_factors(
        narrative_spatial_explicit_diffusion,
        fuel_disagg,
        regions,
        real_values,
        narrative_speed_con_max,
        p_outlier=5.0
    ):
    """
    Arguments
    -------
    spatial_explicit_diffusion : bool
        If regional specific or not
    fuel_disagg : dict
        Disaggregated fuel
    regions : list
        Regions
    real_values : dict
        Real values
    speed_con_max : float
        Maximum diffusion criteria
    p_outlier : float, default=5.0
        Percentage of outliers which are removed from top and bottom
        (Nr of min and max outliers to flatten)

    Returns
    -------
    f_reg : dict
        Spatial diffusion factor
    f_reg_norm : dict
        Spatial diffusion factor
    f_reg_norm_abs : dict
        Spatial diffusion factor
    crit_all_the_same : bool
        Criteria whether regional specific parameters or not
    """
    spatial_explicit_diffusion = narrative_related.read_from_narrative(
        narrative_spatial_explicit_diffusion)

    speed_con_max = narrative_related.read_from_narrative(narrative_speed_con_max)

    if spatial_explicit_diffusion:

        # Define diffusion speed
        speed_con_max = speed_con_max

        # Criteria whether a spatially differentiated diffusion is applied depending on real_values
        crit_all_the_same = False
    else:

        # NEW If not spatial different, set speed_con_max to 1
        speed_con_max = 1

        # Define that the penetration levels are the same for every region
        crit_all_the_same = True

    f_reg, f_reg_norm, f_reg_norm_abs = spatial_diffusion.calc_spatially_diffusion_factors(
        regions=regions,
        fuel_disagg=fuel_disagg,
        real_values=real_values,
        low_congruence_crit=True,
        speed_con_max=speed_con_max,
        p_outlier=p_outlier)

    '''plot_fig_paper = False
    if plot_fig_paper:
        from energy_demand.plotting import result_mapping
        # Global value to distribute
        global_value = 50

        # Select spatial diffusion factor
        #diffusion_vals = f_reg                                  # not weighted
        diffusion_vals = f_reg_norm['ss_space_heating']         # Weighted with enduse
        #diffusion_vals = f_reg_norm_abs['rs_space_heating']    # Absolute distribution (only for capacity installements)

        path_shapefile_input = os.path.abspath(
            'C:/Users/cenv0553/ED/data/_raw_data/C_LAD_geography/lad_2016_uk_simplified.shp')

        result_mapping.plot_spatial_mapping_example(
            diffusion_vals=diffusion_vals,
            global_value=global_value,
            paths=data['result_paths'],
            regions=data['regions'],
            path_shapefile_input=path_shapefile_input,
            plotshow=True)'''

    return f_reg, f_reg_norm, f_reg_norm_abs, crit_all_the_same

def switch_calculations(
        sim_yrs,
        data,
        f_reg,
        f_reg_norm,
        f_reg_norm_abs,
        crit_all_the_same,
        service_switches,
        fuel_switches,
        capacity_switches
    ):
    """This function creates sigmoid diffusion values based
    on defined switches

    Arguments
    ----------
    sim_yrs : list
        Simulated years
    data : dict
        Data container
    f_reg, f_reg_norm, f_reg_norm_abs : str
        Different spatial diffusion values
    crit_all_the_same : dict
        Criteria whether the diffusion is spatial explicit or not

    Returns
    -------
    annual_tech_diff_params : dict
        All values for every simluation year for every technology
    """
    # ---------------------------------------
    # Convert base year fuel input assumptions to energy service for every submodule
    # ---------------------------------------
    s_tech_by_p = {}
    s_fueltype_by_p = {}
    for submodel in data['assumptions'].submodels_names:
        for sector in data['sectors'][submodel]:
            s_tech_by_p[sector], s_fueltype_by_p[sector] = s_fuel_to_service.get_s_fueltype_tech(
                data['enduses'][submodel],
                data['assumptions'].fuel_tech_p_by,
                data['fuels'][submodel],
                data['assumptions'].technologies,
                sector)

    # Get all defined narrative timesteps of all switch types
    narrative_timesteps = {}
    narrative_timesteps.update(get_all_narrative_timesteps(service_switches))
    narrative_timesteps.update(get_all_narrative_timesteps(fuel_switches))
    narrative_timesteps.update(get_all_narrative_timesteps(capacity_switches))

    # ========================================================================================
    # Capacity switches
    #
    # Calculate service shares considering potential capacity installations
    # ========================================================================================

    # Convert globally defined switches to regional switches
    f_diffusion = f_reg_norm_abs # Select diffusion value

    reg_capacity_switches = global_to_reg_capacity_switch(
        data['regions'], capacity_switches, f_diffusion=f_diffusion)

    service_switches_incl_cap = fuel_service_switch.capacity_switch(
        narrative_timesteps,
        data['regions'],
        reg_capacity_switches,
        data['assumptions'].technologies,
        data['fuels']['aggr_sector_fuels'],
        data['assumptions'].fuel_tech_p_by,
        data['assumptions'].base_yr)

    # ======================================================================================
    # Service switches
    #
    # Get service shares of technologies for future year by considering
    # service switches. Potential capacity switches are used as inputs.
    #
    # Autocomplement defined service switches with technologies not
    # explicitly specified in switch on a global scale and distribute spatially
    # in oder that all technologies match up to 100% service share
    # ========================================================================================

    # Select spatial diffusion
    f_diffusion = f_reg_norm

    share_s_tech_ey_p = fuel_service_switch.autocomplete_switches(
        service_switches,
        data['assumptions'].specified_tech_enduse_by,
        s_tech_by_p,
        data['enduses'],
        data['sectors'],
        crit_all_the_same=crit_all_the_same,
        regions=data['regions'],
        f_diffusion=f_diffusion,
        techs_affected_spatial_f=data['assumptions'].techs_affected_spatial_f,
        service_switches_from_capacity=service_switches_incl_cap)

    # ========================================================================================
    # Fuel switches
    #
    # Calculate sigmoid diffusion considering fuel switches and service switches.
    # As inputs, service (and thus also capacity switches) are used
    # ========================================================================================
    diffusion_param_tech = defaultdict(dict)

    for submodel_name in data['assumptions'].submodels_names:
        for enduse in data['enduses'][submodel_name]:
            for sector in data['sectors'][submodel_name]:
                logging.debug("... calculating fuel switches {}  {}  {}".format(submodel_name, enduse, sector))

                diffusion_param_tech[enduse][sector] = sig_param_calc_incl_fuel_switch(
                    narrative_timesteps,
                    data['assumptions'].base_yr,
                    data['assumptions'].crit_switch_happening,
                    data['assumptions'].technologies,
                    enduse=enduse,
                    sector=sector,
                    fuel_switches=fuel_switches,
                    s_tech_by_p=s_tech_by_p[sector][enduse],
                    s_fueltype_by_p=s_fueltype_by_p[sector][enduse],
                    share_s_tech_ey_p=share_s_tech_ey_p,
                    fuel_tech_p_by=data['assumptions'].fuel_tech_p_by[enduse][sector],
                    regions=data['regions'],
                    crit_all_the_same=crit_all_the_same)

    # ------------------
    # Calculate annual values based on calculated parameters for every simulation year
    # ------------------
    annual_tech_diff_params = s_scenario_param.calc_annual_switch_params(
        sim_yrs,
        data['regions'],
        dict(diffusion_param_tech),
        base_yr=data['assumptions'].base_yr,
        s_tech_by_p=s_tech_by_p)

    # ------------------------------------------------------
    # Test whether the calculated service shares sum up to 1
    # and if not, then distribute not assigned demand according to
    # base year distribution
    # Can happen because multiple sigmoid for different technologies
    # ------------------------------------------------------
    for region in annual_tech_diff_params:
        for enduse in annual_tech_diff_params[region]:
            for sector in annual_tech_diff_params[region][enduse]:

                for year in sim_yrs:
                    if annual_tech_diff_params[region][enduse][sector] != []:
                        assigned_service_p = 0
                        for tech in annual_tech_diff_params[region][enduse][sector]:
                            #print("TECH {}  {} {}".format(tech, enduse, sector))
                            assigned_service_p += annual_tech_diff_params[region][enduse][sector][tech][year]

                        # Distribute if not sum up to one according to base year distribution
                        if round(assigned_service_p, 2) != 1:
                            total_not_assined = 1 - assigned_service_p

                            for tech in annual_tech_diff_params[region][enduse][sector]:
                                # Fraction in base year
                                p_by = annual_tech_diff_params[region][enduse][sector][tech][data['assumptions'].base_yr]
                                tech_to_add_not_assigned = total_not_assined * p_by
                                annual_tech_diff_params[region][enduse][sector][tech][year] += tech_to_add_not_assigned

                        assigned_service_p = 0
                        for tech in annual_tech_diff_params[region][enduse][sector]:
                            assigned_service_p += annual_tech_diff_params[region][enduse][sector][tech][year]

                        assert round(assigned_service_p, 2) == 1

    logging.info("Finished calculateding switch_calculations")
    return annual_tech_diff_params

def spatial_explicit_modelling_strategy_vars(
        strategy_vars,
        spatially_modelled_vars,
        regions,
        fuel_disagg,
        f_reg,
        f_reg_norm,
        f_reg_norm_abs
    ):
    """
    Spatial explicit modelling of scenario variables
    based on narratives. Convert strategy variables
    (e.g. country level) to regional variables.

    Arguments
    ---------
    strategy_vars : dict
        All non_spatial strategy variables
    spatially_modelled_vars : dict
        All spatial variables
    regions : dict
        Regions
    fuel_disagg : dict
        Disaggregated fuel
    f_reg, f_reg_norm, f_reg_norm_abs : dict
        Spatial factors
    """
    regional_vars = {}

    # Iterate strategy variables and calculate regional variable
    for var_name, strategy_var in strategy_vars.items():
        logging.debug("...Spatially explicit diffusion modelling %s", var_name)

        single_dim_var = narrative_related.crit_dim_var(strategy_var)

        if single_dim_var:
            regional_vars[var_name] = get_regional_narrative(
                var_name,
                strategy_var,
                spatially_modelled_vars,
                regions,
                fuel_disagg,
                f_reg,
                f_reg_norm,
                f_reg_norm_abs)
        else:
            regional_vars[var_name] = {}

            for sub_var_name, sector_sub_strategy_var in strategy_var.items():
                regional_vars[var_name][sub_var_name] = {}
                if type(sector_sub_strategy_var) is dict: # sectors defined
                    for sector, sub_strategy_var in sector_sub_strategy_var.items():
                        regional_vars[var_name][sub_var_name][sector] = get_regional_narrative(
                            sub_var_name,
                            sub_strategy_var,
                            spatially_modelled_vars,
                            regions,
                            fuel_disagg,
                            f_reg,
                            f_reg_norm,
                            f_reg_norm_abs)
                else:
                    regional_vars[var_name][sub_var_name] = get_regional_narrative(
                        sub_var_name,
                        sector_sub_strategy_var,
                        spatially_modelled_vars,
                        regions,
                        fuel_disagg,
                        f_reg,
                        f_reg_norm,
                        f_reg_norm_abs)

                regional_vars[var_name] = dict(regional_vars[var_name])

    return regional_vars

def get_regional_narrative(
        var_name,
        strategy_vars,
        spatially_modelled_vars,
        regions,
        fuel_disagg,
        f_reg,
        f_reg_norm,
        f_reg_norm_abs
    ):
    """Spatial explicit modelling of scenario variable
    based on narratives. Convert strategy variables
    (e.g. country level) to regional variables.

    Arguments
    ---------
    var_name : str
        Variable name
    strategy_vars : dict
        All non_spatial strategy variables
    spatially_modelled_vars : dict
        All spatial variables
    regions : dict
        Regions
    fuel_disagg : dict
        Disaggregated fuel
    f_reg, f_reg_norm, f_reg_norm_abs : dict
        Spatial factors

    Returns
    ------
    new_narratives : list
        Created narratives
    """
    new_narratives = []
    for narrative in strategy_vars:

        if not narrative['regional_specific']:
            narrative['regional_vals_ey'] = narrative['value_ey']
            narrative['regional_vals_by'] = narrative['value_by']
            new_narratives.append(narrative)
        else:
            regional_vars_by = {}
            regional_vars_ey = {}

            # Check whether scenario variable is regionally modelled
            if var_name not in spatially_modelled_vars:

                # Variable is not spatially modelled
                for region in regions:
                    regional_vars_ey[region] = float(narrative['value_ey'])
                    regional_vars_by[region] = float(narrative['value_by'])
            else:
                if narrative['enduse'] == []:
                    logging.info(
                        "For variable %s no affected enduse is defined. Thus speed is used for diffusion modelling",
                        var_name)
                    fuels_reg = {}
                else:
                    # Get enduse specific fuel for each region
                    fuels_reg = spatial_diffusion.get_enduse_regs(
                        enduse=narrative['enduse'],
                        fuels_disagg=[
                            fuel_disagg['residential'],
                            fuel_disagg['service'],
                            fuel_disagg['industry']])

                # Calculate regional specific strategy variables values for base year
                reg_specific_variables_by = spatial_diffusion.factor_improvements_single(
                    factor_uk=narrative['value_by'],
                    regions=regions,
                    f_reg=f_reg,
                    f_reg_norm=f_reg_norm,
                    f_reg_norm_abs=f_reg_norm_abs,
                    fuel_regs_enduse=fuels_reg)

                # Calculate regional specific strategy variables values for end year
                reg_specific_variables_ey = spatial_diffusion.factor_improvements_single(
                    factor_uk=narrative['value_ey'],
                    regions=regions,
                    f_reg=f_reg,
                    f_reg_norm=f_reg_norm,
                    f_reg_norm_abs=f_reg_norm_abs,
                    fuel_regs_enduse=fuels_reg)

                # Add regional specific strategy variables values
                for region in regions:
                    regional_vars_ey[region] = float(reg_specific_variables_ey[region])
                    regional_vars_by[region] = float(reg_specific_variables_by[region])

            narrative['regional_vals_by'] = regional_vars_by
            narrative['regional_vals_ey'] = regional_vars_ey

            new_narratives.append(narrative)

    return new_narratives

def global_to_reg_capacity_switch(
        regions,
        global_capactiy_switch,
        f_diffusion
    ):
    """Conversion of global capacity switch installations
    to regional installation

    Arguments
    ---------
    regions : list
        Regions
    global_capactiy_switch : float
        Global switch
    f_diffusion : dict

    Returns
    -------
    reg_capacity_switch : dict
        Regional capacity switch
    """
    reg_capacity_switch = {}

    # Get all affected enduses of capacity switches
    switch_enduses = fuel_service_switch.get_all_enduses_of_switches(
        global_capactiy_switch)

    for region in regions:
        reg_capacity_switch[region] = []

        for enduse in switch_enduses:

            # Get all capacity switches related to this enduse
            enduse_capacity_switches = fuel_service_switch.get_switches_of_enduse(
                global_capactiy_switch, enduse, crit_region=False)

            for switch in enduse_capacity_switches:

                global_capacity = switch.installed_capacity
                regional_capacity = global_capacity  * f_diffusion[switch.enduse][region]

                new_switch = read_data.CapacitySwitch(
                    enduse=switch.enduse,
                    technology_install=switch.technology_install,
                    switch_yr=switch.switch_yr,
                    installed_capacity=regional_capacity,
                    sector=switch.sector)

                reg_capacity_switch[region].append(new_switch)

    return reg_capacity_switch

def sum_across_all_submodels_regs(
        regions,
        submodels
    ):
    """Calculate total sum of fuel per region
    """
    fueltypes_nr = lookup_tables.basic_lookups()['fueltypes_nr']
    fuel_aggregated_regs = {}

    for region in regions:
        tot_reg = np.zeros((fueltypes_nr), dtype="float")
        for submodel_fuel_disagg in submodels:
            for fuel_sector_enduse in submodel_fuel_disagg[region].values():

                # Test if fuel for sector
                if isinstance(fuel_sector_enduse, dict):
                    for sector_fuel in fuel_sector_enduse.values():
                        tot_reg += sector_fuel
                else:
                    tot_reg += fuel_sector_enduse

        fuel_aggregated_regs[region] = tot_reg

    return fuel_aggregated_regs

def sum_across_sectors_all_regs(fuel_disagg_reg):
    """Sum fuel across all sectors for every region

    Arguments
    ---------
    fuel_disagg_reg : dict
        Fuel per region, sector and enduse

    Returns
    -------
    fuel_aggregated : dict
        Aggregated fuel per region and enduse
    """
    fuel_aggregated = {}
    for reg, entries in fuel_disagg_reg.items():
        fuel_aggregated[reg] = {}
        for enduse in entries:
            for sector in entries[enduse]:
                fuel_aggregated[reg][enduse] = 0

        for enduse in entries:
            for sector in entries[enduse]:
                fuel_aggregated[reg][enduse] += np.sum(entries[enduse][sector])

    return fuel_aggregated

def sig_param_calc_incl_fuel_switch(
        narrative_timesteps,
        base_yr,
        crit_switch_happening,
        technologies,
        enduse,
        sector,
        fuel_switches,
        s_tech_by_p,
        s_fueltype_by_p,
        share_s_tech_ey_p,
        fuel_tech_p_by,
        regions=False,
        crit_all_the_same=True
    ):
    """Calculate sigmoid diffusion paramaters considering
    fuel and service switches. Test if service switch and
    fuel switch are defined simultaneously (raise error if true).

    Arguments
    ---------
    narrative_timesteps : list
        All defined narrative years
    base_yr : int
        Base year
    crit_switch_happening : dict
        Wheter switch is defined or not per enduse
    technologies : dict
        technologies
    enduse : str
        enduse
    fuel_switches : dict
        fuel switches
    s_tech_by_p : dict
        Service share per technology in base year
    s_fueltype_by_p : dict
        Service share per fueltype for base year
    share_s_tech_ey_p : dict
        Service share per technology for end year
    fuel_tech_p_by : dict
        Fuel share per technology in base year
    regions : list
        Regions
    sector : str
        Sector
    crit_all_the_same : bool,default=True
        Criteria wheter regional specific parameter calculations

    Returns
    -------
    sig_param_tech : dict
        Sigmoid parameters for all affected technologies
    service_switches_out : list
        Service switches
    """
    sig_param_tech = {}
    if sum(s_tech_by_p.values()) == 0: # no fuel is defined for enduse
        logging.debug("no fuel is defined for enduse `{}` and sector `{}`".format(enduse, sector))
    else:
        if share_s_tech_ey_p == {}:
            pass
        else:
            if sector not in list(share_s_tech_ey_p.keys()) or enduse not in list(share_s_tech_ey_p[sector].keys()):
                pass # Nothing defined for this sector or enduse
            else:
                share_s_tech_ey_p = share_s_tech_ey_p[sector][enduse]

        # ----------------------------------------
        # Test if fuel switch is defined for enduse
        # Get affected technologies in fuel switch
        # ----------------------------------------
        _, crit_fuel_switch = s_generate_sigmoid.get_tech_installed(
            enduse, fuel_switches)

        # Test if switch is defined
        crit_switch_service = fuel_service_switch.get_switch_criteria(
            enduse, sector, crit_switch_happening)

        if not crit_switch_service and not crit_fuel_switch:
            pass # no switches defined
        elif crit_switch_service and crit_fuel_switch:
            raise Exception("Error: a fuel and service switch is defined at the same time %s", enduse)
        else:
            any_region = regions[0] # Only calculate for one reg
            switch_yrs = narrative_timesteps[enduse] # Years of narrative

            # Iterate over years defined in narrative
            for switch_yr_cnt, switch_yr in enumerate(switch_yrs):
                sig_param_tech[switch_yr] = {}

                # -------------------------
                # Get starting year of narrative
                # -------------------------
                if switch_yr_cnt == 0: 
                    switch_yr_start = base_yr
                    s_tech_last_narrative_step = s_tech_by_p
                else: # If more than one switch_yr, then take previous year
                    switch_yr_start = switch_yrs[switch_yr_cnt - 1]
                    s_tech_last_narrative_step = s_tech_switched_p[any_region][switch_yr_start]

                # ------------------------------------------
                # Service switch
                #
                # Calculate service shares considering service
                # switches and the diffusion parameters
                # ------------------------------------------
                if crit_switch_service:
                    logging.debug(".... service switch")

                    # Calculate only from service switch
                    s_tech_switched_p = share_s_tech_ey_p

                    # Calculate sigmoid diffusion parameters
                    l_values_sig = s_generate_sigmoid.get_l_values(
                        technologies=technologies,
                        technologies_to_consider=s_tech_last_narrative_step.keys(),
                        regions=regions)
                elif crit_fuel_switch:
                    s_tech_switched_p = defaultdict(dict)
                    l_values_sig = {}
                else:
                    s_tech_switched_p = defaultdict(dict)

                # ------------------------------------------
                # Fuel switch
                # ------------------------------------------
                if crit_fuel_switch:
                    logging.info(".... fuel switch")
                    """Calculate future service share after fuel switches
                    and calculte sigmoid diffusion paramters."""
                    # Get fuel switches of enduse and switch_yr
                    enduse_fuel_switches = fuel_service_switch.get_switches_of_enduse(
                        fuel_switches, enduse, switch_yr, crit_region=False)

                    if crit_all_the_same:
                        logging.debug("... calculating fuel switches (not regional specific): {}".format(enduse))
                        # Calculate service demand after fuel switches for each technology
                        s_tech_switched_p_values_all_regs = s_generate_sigmoid.calc_service_fuel_switched(
                            enduse_fuel_switches,
                            technologies,
                            s_fueltype_by_p,
                            s_tech_last_narrative_step,
                            fuel_tech_p_by,
                            'actual_switch')

                        # Calculate L for every technology for sigmod diffusion
                        l_values_all_regs = s_generate_sigmoid.tech_l_sigmoid(
                            s_tech_switched_p_values_all_regs,
                            enduse_fuel_switches,
                            technologies,
                            s_tech_last_narrative_step.keys(),
                            s_fueltype_by_p,
                            s_tech_last_narrative_step,
                            fuel_tech_p_by)

                        for region in regions:
                            s_tech_switched_p[region][switch_yr] = s_tech_switched_p_values_all_regs
                            l_values_sig[region] = l_values_all_regs

                        s_tech_switched_p = dict(s_tech_switched_p)
                    else:
                        logging.debug("... calculating fuel switches (regional specific): {}".format(enduse))
                        for region in regions:

                            # Calculate service demand after fuel switches for each technology
                            s_tech_switched_p[region][switch_yr] = s_generate_sigmoid.calc_service_fuel_switched(
                                enduse_fuel_switches,
                                technologies,
                                s_fueltype_by_p,
                                s_tech_last_narrative_step,
                                fuel_tech_p_by,
                                'actual_switch')

                            # Calculate L for every technology for sigmod diffusion
                            l_values_sig[region] = s_generate_sigmoid.tech_l_sigmoid(
                                s_tech_switched_p[region][switch_yr],
                                enduse_fuel_switches,
                                technologies,
                                s_tech_last_narrative_step.keys(),
                                s_fueltype_by_p,
                                s_tech_last_narrative_step,
                                fuel_tech_p_by)

                        s_tech_switched_p = dict(s_tech_switched_p)

                # -----------------------------------------------
                # Calculates parameters for sigmoid diffusion of
                # technologies which are switched to/installed.
                # -----------------------------------------------
                if crit_all_the_same:
                    logging.debug("... calc parameters of `{}` for year `{}`  {}".format(enduse, switch_yr, sector))
                    sig_param_tech_all_regs_value = s_generate_sigmoid.tech_sigmoid_parameters(
                        switch_yr,
                        switch_yr_start,
                        technologies,
                        l_values_sig[any_region],
                        s_tech_last_narrative_step,
                        s_tech_switched_p[any_region][switch_yr])

                    for region in regions:
                        sig_param_tech[switch_yr][region] = sig_param_tech_all_regs_value
                else:
                    logging.debug("... calc region specific parameters of `{}` for year `{}` sector: {}".format(enduse, switch_yr, sector))
                    for region in regions:
                        sig_param_tech[switch_yr][region] = s_generate_sigmoid.tech_sigmoid_parameters(
                            switch_yr,
                            switch_yr_start,
                            technologies,
                            l_values_sig[region],
                            s_tech_by_p,
                            s_tech_switched_p[region][switch_yr])

    return sig_param_tech

def get_sector__enduse_switches(
        sector_to_match,
        enduse_to_match,
        switches
    ):
    """Get all switches of a sector if the switches are
    defined specifically for a sector. If the switches are
    not specifically for a sector, return all switches

    Arguments
    ----------
    sector_to_match : int
        Sector to find switches
    switches : list
        Switches

    Returns
    -------
    switches : list
        Switches of sector
    """
    # Get all sectors for this enduse
    switches_out = set([])

    for switch in switches:
        if switch.sector == sector_to_match and switch.enduse == enduse_to_match:
            switches_out.add(switch)
        elif not switch.sector and switch.enduse == enduse_to_match: # Not defined specifically for sectors and add all
            switches_out.add(switch)
        else:
            pass

    return list(switches_out)



"""File to read in all service sector related data
"""
import os
import sys
import csv
from datetime import date
import numpy as np
from energy_demand.read_write import read_data
from energy_demand.read_write import write_data
from energy_demand.basic import date_prop
from energy_demand.profiles import load_profile

def dict_init_carbon_trust():
    """Helper function to initialise dict
    """
    carbon_trust_raw = {}
    for day in range(365):
        day_dict_h = {k: [] for k in range(24)}
        carbon_trust_raw[day] = day_dict_h

    return carbon_trust_raw

def read_raw_carbon_trust_data(folder_path):
    """Read in raw carbon trust dataset (used for service sector)

    Arguments
    ----------
    foder_path : string
        Path to folder with stored csv files

    Returns
    -------
    load_shape_y_dh : array
        Load shape for every day (tot sum 365) ((365, 24))
    load_peak_shape_dh : array
        Peak loadshape for peak day ((24))
    shape_non_peak_yd : array
        Yh load profile ((365))

    Note
    -----
    1. Get gas peak day load shape (the max daily demand can be taken from weather data,
       the daily shape however is not provided by samson)
    2. Iterate individual files which are about a year (even though gaps exist)
    3. Select those day with the maximum load
    4. Get the hourly shape of this day
    5. Calculate total demand of every day
    6. Assign percentag of total daily demand to each hour
    """
    def initialise_main_dict():
        """Helper function to initialise dict
        """
        out_dict_av = {'working_day': {}, 'holiday': {}}
        for dtype in out_dict_av:
            month_dict = {}
            for month in range(12):
                month_dict[month] = {k: [] for k in range(24)}
            out_dict_av[dtype] = month_dict
        return out_dict_av

    def initialise_out_dict_av():
        """Helper function to initialise dict"""
        out_dict_av = {'working_day': {}, 'holiday': {}}
        for dtype in out_dict_av:
            month_dict = {}
            for month in range(12):
                month_dict[month] = {k: 0 for k in range(24)}
            out_dict_av[dtype] = month_dict
        return out_dict_av

    # Get all files in folder
    all_csv_in_folder = os.listdir(folder_path)
    main_dict = initialise_main_dict()
    carbon_trust_raw = dict_init_carbon_trust()

    nr_of_line_entries = 0
    dict_max_dh_shape = {}

    # Itreatu folder with csv files
    for path_csv_file in all_csv_in_folder:
        path_csv_file = os.path.join(folder_path, path_csv_file)

        # Read csv file
        with open(path_csv_file, 'r') as csv_file:
            read_lines = csv.reader(csv_file, delimiter=',')
            _headings = next(read_lines)
            max_d_demand = 0 # Used for searching maximum

            # Count number of lines in CSV file
            row_data = []
            for count_row, row in enumerate(read_lines):
                row_data.append(row)
            #print("Number of lines in csv file: " + str(count_row))

            # Calc yearly demand based on one year data measurements
            if count_row > 365: # if more than one year is in csv file
                #print("FILE covers a full year---------------------------")

                # Test if file has correct form and not more entries than 48 half-hourly entries
                for day, row in enumerate(row_data):
                    if len(row) != 49:
                        continue # Skip row

                    # Use only data of one year
                    if day > 365:
                        continue

                    load_shape_dh = np.zeros((24), dtype="float")

                    row[1:] = map(float, row[1:]) # Convert all values except date into float values
                    daily_sum = sum(row[1:]) # Total daily sum
                    nr_of_line_entries += 1 # Nr of lines added
                    day = int(row[0].split("/")[0])
                    month = int(row[0].split("/")[1])
                    year = int(row[0].split("/")[2])

                    # Redefine yearday to another year and skip 28. of Feb.
                    if date_prop.is_leap_year(int(year)) is True:
                        year = year + 1 # Shift whole dataset to another year
                        if month == 2 and day == 29:
                            continue #skip leap day

                    date_row = date(year, month, day)
                    daytype = date_prop.get_weekday_type(date_row)

                    yearday_python = date_row.timetuple().tm_yday - 1 # - 1 because in _info: 1.Jan = 1
                    month_python = month - 1 # Month Python

                    h_day, cnt, control_sum = 0, 0, 0

                    # -----------------------------------------------
                    # Iterate half hour data and summarise to hourly
                    # -----------------------------------------------
                    for data_h in row[1:]:  # Skip first date row in csv file
                        cnt += 1
                        if cnt == 2:
                            demand_h = first_data_h + data_h
                            control_sum += abs(demand_h)

                            # Add demand
                            carbon_trust_raw[yearday_python][h_day].append(demand_h)

                            # Store demand according to daytype (aggregated by doing so)
                            main_dict[daytype][month_python][h_day].append(demand_h)

                            if daily_sum == 0: # Skip row if no demand of the day
                                load_shape_dh[h_day] = 0
                                continue
                            else:
                                load_shape_dh[h_day] = demand_h / daily_sum

                            cnt = 0
                            h_day += 1

                        # Value lagging behind one iteration
                        first_data_h = data_h

                    # Test if this is the day with maximum demand of this CSV file
                    if daily_sum >= max_d_demand:
                        max_d_demand = daily_sum
                        max_dh_shape = load_shape_dh

                    # Check if 100 %
                    np.testing.assert_almost_equal(control_sum, daily_sum, decimal=7, err_msg="")

                # Add load shape of maximum day in csv file
                dict_max_dh_shape[path_csv_file] = max_dh_shape

    # ---------------
    # Data processing
    # ---------------
    # --Average average maxium peak dh of every csv file
    load_peak_average_dh = np.zeros((24), dtype="float")
    for peak_shape_dh in dict_max_dh_shape.values():
        load_peak_average_dh += peak_shape_dh
    load_peak_shape_dh = load_peak_average_dh / len(dict_max_dh_shape)

    # -----------------------------------------------
    # Calculate average load shapes for every month
    # -----------------------------------------------
    out_dict_av = initialise_out_dict_av()

    for daytype in main_dict:
        for month in main_dict[daytype]:
            for hour in main_dict[daytype][month]:
                nr_of_entries = len(main_dict[daytype][month][hour])
                if nr_of_entries != 0:
                    out_dict_av[daytype][month][hour] = sum(main_dict[daytype][month][hour]) / nr_of_entries

    # ----------------------------------------------------------
    # Distribute raw data into base year depending on daytype
    # ----------------------------------------------------------
    year_data = assign_data_to_year(out_dict_av, 2015)

    # Calculate yearly sum
    yearly_demand = np.sum(year_data)

    # Create load_shape_dh
    load_shape_y_dh = np.zeros((365, 24), dtype="float")
    for day, dh_values in enumerate(year_data):
        load_shape_y_dh[day] = load_profile.abs_to_rel(dh_values) # daily shape

    np.testing.assert_almost_equal(np.sum(load_shape_y_dh), 365, decimal=2, err_msg="")

    # Calculate shape_non_peak_yd
    shape_non_peak_yd = np.zeros((365), dtype="float")
    for yearday, carbon_trust_d in enumerate(year_data):
        shape_non_peak_yd[yearday] = np.sum(carbon_trust_d)

    shape_non_peak_yd = shape_non_peak_yd / yearly_demand

    np.testing.assert_almost_equal(np.sum(shape_non_peak_yd), 1, decimal=2, err_msg="")

    return load_shape_y_dh, load_peak_shape_dh, shape_non_peak_yd

def assign_data_to_year(carbon_trust_data, base_yr):
    """Fill every base year day with correct data

    Arguments
    ----------
    carbon_trust_data : data
        Raw data
    base_yr : int
        Base Year
    """
    shape_non_peak_y_dh = np.zeros((365, 24), dtype="float")

    # Create list with all dates of a whole year
    list_dates = date_prop.fullyear_dates(
        start=date(base_yr, 1, 1),
        end=date(base_yr, 12, 31))

    # Assign every date to the place in the array of the year
    for yearday in list_dates:
        month_python = yearday.timetuple().tm_mon - 1 # - 1 because in _info: Month 1 = Jan
        yearday_python = yearday.timetuple().tm_yday - 1 # - 1 because in _info: 1.Jan = 1
        daytype = date_prop.get_weekday_type(yearday)

        # Get day from HES raw data array
        _data = carbon_trust_data[daytype][month_python]

        # Add values to yearly
        _data = np.array(list(_data.items()))
        shape_non_peak_y_dh[yearday_python] = np.array(_data[:, 1], dtype="float")

    return shape_non_peak_y_dh

def run(paths, local_paths, lookups):
    """Function to run script
    """
    print("... start script %s", os.path.basename(__file__))
    _, ss_sectors, ss_enduses = read_data.read_fuel_ss(
        paths['ss_fuel_raw'],
        lookups['fueltypes_nr'])

    # Iterate sectors and read in shape
    for sector in ss_sectors:
        # Match electricity shapes for every sector
        # to correct folder with load profiles
        if sector == 'community_arts_leisure':
            sector_folder_path_elec = os.path.join(
                local_paths['folder_raw_carbon_trust'], "Community")
        elif sector == 'education':
            sector_folder_path_elec = os.path.join(
                local_paths['folder_raw_carbon_trust'], "Education")
        elif sector == 'emergency_services':
            sector_folder_path_elec = os.path.join(
                local_paths['folder_raw_carbon_trust'], "_all_elec")
        elif sector == 'health':
            sector_folder_path_elec = os.path.join(
                local_paths['folder_raw_carbon_trust'], "Health")
        elif sector == 'hospitality':
            sector_folder_path_elec = os.path.join(
                local_paths['folder_raw_carbon_trust'], "_all_elec")
        elif sector == 'military':
            sector_folder_path_elec = os.path.join(
                local_paths['folder_raw_carbon_trust'], "_all_elec")
        elif sector == 'offices':
            sector_folder_path_elec = os.path.join(
                local_paths['folder_raw_carbon_trust'], "Offices")
        elif sector == 'retail':
            sector_folder_path_elec = os.path.join(
                local_paths['folder_raw_carbon_trust'], "Retail")
        elif sector == 'storage':
            sector_folder_path_elec = os.path.join(
                local_paths['folder_raw_carbon_trust'], "_all_elec")
        else:
            raise Exception("Error: The sector {} could not be assigned".format(sector))

        # ------------------------------------------------------
        # Assign shape across enduse for service sector
        # ------------------------------------------------------
        for enduse in ss_enduses:
            print("Enduse service: %s in sector %s", enduse, sector)

            # Enduses
            enduses_mainly_gas = ['ss_water_heating', 'ss_space_heating', 'ss_other_gas']

            # Select shape depending on enduse
            if enduse in enduses_mainly_gas:
                folder_path = os.path.join(
                    local_paths['folder_raw_carbon_trust'],
                    "_all_gas")
            else:
                folder_path = sector_folder_path_elec

            # Read in shape from carbon trust metering trial dataset
            shape_non_peak_y_dh, load_peak_shape_dh, shape_non_peak_yd = read_raw_carbon_trust_data(
                folder_path)

            # Write shapes to txt
            joint_string_name = str(sector) + "__" + str(enduse)

            write_data.create_txt_shapes(
                joint_string_name,
                local_paths['ss_load_profile_txt'],
                load_peak_shape_dh,
                shape_non_peak_y_dh,
                shape_non_peak_yd)

    print("... finished script %s", os.path.basename(__file__))
    return



"""
"""
from energy_demand.geography import weather_region
from energy_demand.assumptions import strategy_vars_def
import numpy as np

'''def test_WeatherRegion():

    from energy_demand.read_write import data_loader
    from energy_demand.assumptions import general_assumptions
    path_main = os.path.abspath("C://Users//cenv0553//nismod//models//energy_demand")
    path_main = os.path.join("")

    # Load data
    data = {}
    data['paths'] = data_loader.load_paths(path_main)
    data['local_paths'] = data_loader.get_local_paths(path_main)
    data['lookups'] = lookup_tables.basic_lookups()
    data['enduses'], data['sectors'], data['fuels'] = data_loader.load_fuels(data['paths'], data['lookups'])
    
    #{'base_yr': base_yr, 'curr_yr': curr_yr, 'sim_period_yrs': sim_period_yrs}
    #Load assumptions


    strategy_vars_def.load_param_assump(data['paths'], data['enduses'])
    data['assumptions'] = read_data.read_param_yaml(data['paths']['yaml_parameters'])
    data['tech_lp'] = data_loader.load_data_profiles(data['paths'], data['local_paths'], data['assumptions'])

    weather_region_obj = weather_region.WeatherRegion(
        weather_region_name="Bern",
        sim_param=data['sim_param'],
        assumptions=data['assumptions'],
        lookups=data['lookups'],
        enduses=['heating', 'cooking'],
        temp_by=np.zeros((365, 24)),
        tech_lp=data['tech_lp'],
        sectors=["sec_A"]
    )
    assert weather_region_obj.weather_region_name == 'Bern'
test_WeatherRegion()'''



"""
Testing s_generate_sigmoid
"""
import numpy as np
from energy_demand.scripts import s_generate_sigmoid
from energy_demand.technologies import diffusion_technologies
from energy_demand.read_write import read_data
from energy_demand.technologies import tech_related

def test_tech_sigmoid_paramters():
    """testing
    """

    fuel_switches = [read_data.FuelSwitch(
        enduse='heating',
        technology_install='boilerA',
        switch_yr=2050)]

    technologies = {
        'boilerA': read_data.TechnologyData(market_entry=1990),
        'boilerB': read_data.TechnologyData(market_entry=1990)}

    assump_fy = 1.0
    result = s_generate_sigmoid.tech_sigmoid_parameters(
        yr_until_switched=2050,
        switch_yr_start=2010,
        technologies=technologies,
        l_values={'boilerA': 1.0, 'boilerB': 1.0},
        s_tech_by_p={'boilerA': 0.5, 'boilerB': 0.5},
        s_tech_switched_p={'boilerA': assump_fy, 'boilerB': 0})

    y_calculated = diffusion_technologies.sigmoid_function(
        2050, 1.0, result['boilerA']['midpoint'], result['boilerA']['steepness'])

    assert y_calculated >= (assump_fy - 0.02) and y_calculated <= assump_fy + 0.02

    # ------------
    assump_fy = 1.0
    result = s_generate_sigmoid.tech_sigmoid_parameters(
        yr_until_switched=2050,
        switch_yr_start=2010,
        technologies=technologies,
        l_values={'boilerA': 1.0, 'boilerB': 1.0},
        s_tech_by_p={'boilerA': 0.5, 'boilerB': 0.5},
        s_tech_switched_p={'boilerA': assump_fy, 'boilerB': 0})

    y_calculated = diffusion_technologies.sigmoid_function(
        2050, 1.0, result['boilerA']['midpoint'], result['boilerA']['steepness'])

    assert y_calculated >= (assump_fy - 0.02) and y_calculated <= assump_fy + 0.02

def test_calc_sigmoid_parameters():
    """Testing
    """
    l_value = 0.5
    xdata = np.array([2020.0, 2050.0])
    ydata = np.array([0.1, 0.2]) #[point_y_by, point_y_projected]

    # fit parameters
    fit_parameter = s_generate_sigmoid.calc_sigmoid_parameters(
        l_value,
        xdata,
        ydata)

    y_calculated = diffusion_technologies.sigmoid_function(xdata[1], l_value, *fit_parameter)

    assert round(y_calculated, 3) == round(ydata[1], 3)

def test_calc_sigmoid_parameters2():
    """Testing
    """
    l_value = 1.0
    xdata = np.array([2020.0, 2060.0])
    ydata = np.array([0.01, 0.99])

    # fit parameters
    fit_parameter = s_generate_sigmoid.calc_sigmoid_parameters(
        l_value,
        xdata,
        ydata)

    y_calculated = diffusion_technologies.sigmoid_function(xdata[1], l_value, *fit_parameter)

    assert round(y_calculated, 3) == round(ydata[1], 3)

def test_get_tech_installed():
    """"""
    enduse = 'heating'
    fuel_switches = [
        read_data.FuelSwitch(
            enduse='heating',
            technology_install='boilerB'
        ),
        read_data.FuelSwitch(
            enduse='heating',
            technology_install='boilerA'),
        read_data.FuelSwitch(
            enduse='cooking',
            technology_install='techC'
        )
        ]

    result = s_generate_sigmoid.get_tech_installed(enduse, fuel_switches)

    expected = {'heating': ['boilerB', 'boilerA']}

    assert 'boilerA' in expected['heating']
    assert 'boilerB' in expected['heating']

def test_calc_service_fuel_switched():
    """
    """
    fueltype_lookup = {
        'solid_fuel': 0,
        'gas': 1,
        'electricity': 2,
        'oil': 3,
        'heat_sold': 4,
        'biomass': 5,
        'hydrogen': 6,
        'heat': 7}

    technologies = {
        'boilerA': read_data.TechnologyData(
            fueltype='gas',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990),
        'boilerB': read_data.TechnologyData(
            fueltype='electricity',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990),
        'boilerC': read_data.TechnologyData(
            fueltype='gas',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990)}

    enduse = 'heating'

    fuel_switches = [
        read_data.FuelSwitch(
            enduse='heating',
            technology_install='boilerB',
            switch_yr=2020,
            fueltype_replace=tech_related.get_fueltype_int('gas'),
            fuel_share_switched_ey=1.0
        )]

    service_fueltype_p = {1: 1.0, 2: 0.0}
    s_tech_by_p = {'boilerA': 1.0, 'boilerB': 0.0}
    fuel_tech_p_by = {1: {'boilerA': 1.0}, 2: {'boilerB': 1.0}}

    result = s_generate_sigmoid.calc_service_fuel_switched(
        fuel_switches,
        technologies,
        service_fueltype_p,
        s_tech_by_p,
        fuel_tech_p_by,
        switch_type='actual_switch')

    assert result['boilerB'] == 1.0
    assert result['boilerA'] == 0.0

    # -------

    fuel_switches = [
        read_data.FuelSwitch(
            enduse='heating',
            technology_install='boilerB',
            switch_yr=3050,
            fueltype_replace=tech_related.get_fueltype_int('gas'),
            fuel_share_switched_ey=0.5
        )]

    service_fueltype_p = {1: 1.0, 2: 0.0}
    s_tech_by_p = {'boilerA': 1.0, 'boilerB': 0.0}
    fuel_tech_p_by = {1: {'boilerA': 1.0}, 2: {'boilerB': 1.0}}

    result = s_generate_sigmoid.calc_service_fuel_switched(
        fuel_switches,
        technologies,
        service_fueltype_p,
        s_tech_by_p,
        fuel_tech_p_by,
        switch_type='actual_switch')

    assert result['boilerB'] == 0.5
    assert result['boilerA'] == 0.5

    # -------

    fuel_switches = [
        read_data.FuelSwitch(
            enduse='heating',
            technology_install='boilerB',
            switch_yr=3050,
            fueltype_replace=tech_related.get_fueltype_int('gas'),
            fuel_share_switched_ey=0.5
        )]

    service_fueltype_p = {1: 0.5, 2: 0.5}
    s_tech_by_p = {'boilerA': 0.5, 'boilerB': 0.5}
    fuel_tech_p_by = {1: {'boilerA': 1.0}, 2: {'boilerB': 1.0}}

    result = s_generate_sigmoid.calc_service_fuel_switched(
        fuel_switches,
        technologies,
        service_fueltype_p,
        s_tech_by_p,
        fuel_tech_p_by,
        switch_type='actual_switch')

    assert result['boilerB'] == 0.75
    assert result['boilerA'] == 0.25

    # -------

    fuel_switches = [
        read_data.FuelSwitch(
            enduse='heating',
            technology_install='boilerB',
            switch_yr=3050,
            fueltype_replace=tech_related.get_fueltype_int('gas'),
            fuel_share_switched_ey=0.5
        )]

    service_fueltype_p = {1: 0.5, 2: 0.5}
    s_tech_by_p = {'boilerA': 0.25, 'boilerB': 0.5, 'boilerC': 0.25}
    fuel_tech_p_by  = {1: {'boilerA': 0.5, 'boilerC': 0.5,}, 2: {'boilerB': 1.0}}

    result = s_generate_sigmoid.calc_service_fuel_switched(
        fuel_switches,
        technologies,
        service_fueltype_p,
        s_tech_by_p,
        fuel_tech_p_by,
        switch_type='actual_switch')

    assert result['boilerC'] == 0.125
    assert result['boilerB'] == 0.75
    assert result['boilerA'] == 0.125

def test_tech_l_sigmoid():
    """testing
    """
    fueltype_lookup = {
        'solid_fuel': 0,
        'gas': 1,
        'electricity': 2,
        'oil': 3,
        'heat_sold': 4,
        'biomass': 5,
        'hydrogen': 6,
        'heat': 7}

    technologies = {
        'boilerA': read_data.TechnologyData(
            fueltype='gas',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=1.0),

        'boilerB': read_data.TechnologyData(
            fueltype='electricity',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=1.0)
        }

    fuel_switches = [
        read_data.FuelSwitch(
            enduse='heating',
            technology_install='boilerB',
            switch_yr=2020,
            fueltype_replace=tech_related.get_fueltype_int('gas'),
            fuel_share_switched_ey=1.0
        )]

    service_fueltype_p =  {1: 1.0, 2: 0.0}
    s_tech_by_p = {
        'boilerA': 1.0,
        'boilerB': 0.0}

    fuel_tech_p_by = {
        1: {'boilerA': 1.0},
        2: {'boilerB': 1.0}}
    installed_tech = ['boilerB']

    s_tech_switched_ey = {
        'boilerA': 0.0,
        'boilerB': 1.0}

    result = s_generate_sigmoid.tech_l_sigmoid(
        s_tech_switched_ey=s_tech_switched_ey,
        enduse_fuel_switches=fuel_switches,
        technologies=technologies,
        installed_tech=installed_tech,
        s_fueltype_by_p=service_fueltype_p,
        s_tech_by_p=s_tech_by_p,
        fuel_tech_p_by=fuel_tech_p_by)

    assert result['boilerB'] == 1.0

    # -----
    technologies = {
        'boilerA': read_data.TechnologyData(
            fueltype='gas',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=1.0),

        'boilerB': read_data.TechnologyData(
            fueltype='electricity',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=0.8)
        }

    fuel_switches = [
        read_data.FuelSwitch(
            enduse='heating',
            technology_install='boilerB',
            switch_yr=2020,
            fueltype_replace=tech_related.get_fueltype_int('gas'),
            fuel_share_switched_ey=0.5 #info lower than max
        )]

    service_fueltype_p = {1: 1.0, 2: 0.0}
    s_tech_by_p = {'boilerA': 1.0, 'boilerB': 0.0}
    fuel_tech_p_by = {1: {'boilerA': 1.0}, 2: {'boilerB': 1.0}}
    installed_tech = ['boilerB']

    s_tech_switched_ey = {'boilerA': 0.5, 'boilerB': 0.5}

    result = s_generate_sigmoid.tech_l_sigmoid(
        s_tech_switched_ey=s_tech_switched_ey,
        enduse_fuel_switches=fuel_switches,
        technologies=technologies,
        installed_tech=installed_tech,
        s_fueltype_by_p=service_fueltype_p,
        s_tech_by_p=s_tech_by_p,
        fuel_tech_p_by=fuel_tech_p_by)
    assert result['boilerB'] == 0.8

def test_get_sig_diffusion():
    """
    """
    fueltype_lookup = {
        'solid_fuel': 0,
        'gas': 1,
        'electricity': 2,
        'oil': 3,
        'heat_sold': 4,
        'biomass': 5,
        'hydrogen': 6,
        'heat': 7}

    technologies = {
        'boilerA': read_data.TechnologyData(
            fueltype='gas',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=1.0),
        'boilerC': read_data.TechnologyData(
            fueltype='gas',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=0.999),
        'boilerB': read_data.TechnologyData(
            fueltype='electricity',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=1.0)}

    tech_increased_service = ['boilerA']

    regions = ['regA']

    sig_param = s_generate_sigmoid.get_l_values( 
        technologies,
        tech_increased_service,
        regions)

    assert sig_param['regA']['boilerA'] == 1.0

    # -----

    tech_increased_service = ['boilerC']

    sig_param = s_generate_sigmoid.get_l_values(
        technologies,
        tech_increased_service,
        regions)

    assert sig_param['regA']['boilerC'] == 0.999

    # -----

    tech_increased_service = ['boilerC']

    sig_param = s_generate_sigmoid.get_l_values(
        technologies,
        tech_increased_service,
        regions=regions)

    assert sig_param['regA']['boilerC'] == 0.999



"""Plot map of shapefile
"""
# pylint: disable=C0103
import argparse
import os

import cartopy.crs as ccrs
import cartopy.io.shapereader as shpreader
import matplotlib.pyplot as plt

def plot_shapes(path):
    plt.figure(figsize=(6, 6), dpi=150)
    proj = ccrs.OSGB()
    ax = plt.axes(projection=proj)
    ax.outline_patch.set_visible(False)

    geoms = []
    for record in shpreader.Reader(path).records():
        geoms.append(record.geometry)

    ax.add_geometries(geoms, crs=proj, edgecolor='white', facecolor='#efefef')

    output_filename = os.path.join(
        os.getcwd(),
        'map.png'
    )
    plt.savefig(output_filename)


def add_geoms_by_attribute(path, ax):
    """Example adding geometries with categorical color
    """
    # set up a dict to hold geometries keyed by our key
    geoms_by_key = defaultdict(list)
    # here we hardcode colors by key - could use other methods (interpolation, lookup...)
    colors_by_key = {
        'a': '#ff0000',
        'b': '#00ff00',
        'c': '#0000ff'
    }

    # for each records, pick out our key's value from the record
    # and store the geometry in the relevant list under geoms_by_key
    for record in shpreader.Reader(path).records():
        key = record.attributes['key']
        geoms_by_key[key].append(record.geometry)

    # now we have all the geometries in lists for each value of our key
    # add them to the axis, using the relevant color as facecolor
    for key, geoms in geoms_by_key:
        color = colors_by_key[key]
        ax.add_geometries(geoms, crs=proj, edgecolor='white', facecolor=color)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Plot a shapefile.')
    parser.add_argument('path', help='path to the shapefile to plot')
    args = parser.parse_args()

    plot_shapes(args.path)